# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Júnior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'
# This script aims to be a post installation configuration script, setting
# shortcuts, plugins, ...

import sys
from . import error

if sys.platform.startswith('win32'):
    # Create the functions to deal with Windows registry, from http://stackoverflow.com/a/35286642
    if sys.version_info < (3, 0):
        from _winreg import (OpenKey, HKEY_CURRENT_USER, ConnectRegistry, KEY_READ, KEY_WOW64_64KEY, EnumValue, CloseKey, EnumKey, QueryValueEx, REG_SZ,
                             CreateKey, KEY_WRITE, SetValueEx, DeleteKey)
        OpenKeyError = WindowsError
        ConnectRegistryError = WindowsError
    else:
        from winreg import (OpenKey, HKEY_CURRENT_USER, ConnectRegistry, KEY_READ, KEY_WOW64_64KEY, EnumValue, CloseKey, EnumKey, QueryValueEx, REG_SZ,
                            CreateKey, KEY_WRITE, SetValueEx, DeleteKey)
        OpenKeyError = FileNotFoundError  # noqa: F821
        ConnectRegistryError = PermissionError  # noqa: F821
    __all__ = ['reg_enum_values', 'reg_enum_keys', 'reg_get', 'reg_set', 'reg_del']

    def reg_enum_values(path, key=HKEY_CURRENT_USER):
        # Read variable from Windows Registry.
        try:
            reg = ConnectRegistry(None, key)
            try:
                registry_key = OpenKey(reg, path, 0, KEY_READ)
            except OpenKeyError:
                registry_key = OpenKey(reg, path, 0, KEY_READ | KEY_WOW64_64KEY)
            values = ()
            try:
                idx = 0
                while 1:
                    values.append(EnumValue(registry_key, idx))
                    idx = idx + 1
            except WindowsError:
                pass
            return values
            # TODO this portion of code was unreachable, remove?
            # CloseKey(reg)
            # return value
        except WindowsError:
            pass
        return None

    def reg_enum_keys(path, key=HKEY_CURRENT_USER):
        # Read variable from Windows Registry.
        try:
            reg = ConnectRegistry(None, key)
            try:
                registry_key = OpenKey(reg, path, 0, KEY_READ)
            except OpenKeyError:
                registry_key = OpenKey(reg, path, 0, KEY_READ | KEY_WOW64_64KEY)
            sub_keys = []
            try:
                idx = 0
                while 1:
                    sub_keys.append(EnumKey(registry_key, idx))
                    idx = idx + 1
            except WindowsError:
                pass
            CloseKey(reg)
            return sub_keys
        except WindowsError:
            pass
        return None

    def reg_get(path, name, key=HKEY_CURRENT_USER):
        # Read variable from Windows Registry.
        try:
            reg = ConnectRegistry(None, key)
            try:
                registry_key = OpenKey(reg, path, 0, KEY_READ)
            except OpenKeyError:
                registry_key = OpenKey(reg, path, 0, KEY_READ | KEY_WOW64_64KEY)
            value, regtype = QueryValueEx(registry_key, name)
            CloseKey(reg)
            return value
        except WindowsError:
            pass
        return None

    def reg_set(path, name, value, key=HKEY_CURRENT_USER, key_type=REG_SZ):
        # Write in the Windows Registry.
        try:
            reg = ConnectRegistry(None, key)
            CreateKey(reg, path)
            try:
                registry_key = OpenKey(reg, path, 0, KEY_WRITE)
            except OpenKeyError:
                registry_key = OpenKey(reg, path, 0, KEY_WRITE | KEY_WOW64_64KEY)
            SetValueEx(registry_key, name, 0, key_type, value)
            CloseKey(reg)
            # Update the Windows behaviour.
            # SendMessage(win32con.HWND_BROADCAST, win32con.WM_SETTINGCHANGE, 0, 'Environment')
            return True
        except ConnectRegistryError:
            error('You should run this command as system administrator: run the terminal as administrator and type the command again.')
        except WindowsError:
            pass
        return False

    def reg_del(name, key=HKEY_CURRENT_USER):
        # Delete a registry key on Windows.
        try:
            reg = ConnectRegistry(None, key)
            # registry_key = OpenKey(reg, name_base, 0, KEY_ALL_ACCESS)
            DeleteKey(reg, name)
            CloseKey(reg)
            # Update the Windows behaviour.
            # SendMessage(win32con.HWND_BROADCAST, win32con.WM_SETTINGCHANGE, 0, 'Environment')
            return True
        except ConnectRegistryError:
            error('You should run this command as system administrator: run the terminal as administrator and type the command again.')
        except WindowsError:
            pass
        return False



#!/usr/bin/python3
# -*- coding: utf-8 -*-
currency_symbols = {'AUD': 'A$',
                    'BGN': 'BGN',
                    'BRL': 'R$',
                    'CAD': 'CA$',
                    'CHF': 'CHF',
                    'CNY': 'CN¥',
                    'CZK': 'CZK',
                    'DKK': 'DKK',
                    'EUR': '€',
                    'GBP': '£',
                    'HKD': 'HK$',
                    'HRK': 'HRK',
                    'HUF': 'HUF',
                    'IDR': 'IDR',
                    'ILS': '₪',
                    'INR': '₹',
                    'ISK': 'ISK',
                    'JPY': '¥',
                    'KRW': '₩',
                    'MXN': 'MX$',
                    'MYR': 'MYR',
                    'NOK': 'NOK',
                    'NZD': 'NZ$',
                    'PHP': 'PHP',
                    'PLN': 'PLN',
                    'RON': 'RON',
                    'SEK': 'SEK',
                    'SGD': 'SGD',
                    'THB': 'THB',
                    'TRY': 'TRY',
                    'USD': '$',
                    'ZAR': 'ZAR',
                    }
currency_names = {'AUD': 'Australian Dollar',
                  'BGN': 'Bulgarian Lev',
                  'BRL': 'Brazilian Real',
                  'CAD': 'Canadian Dollar',
                  'CHF': 'Swiss Franc',
                  'CNY': 'Chinese Yuan',
                  'CZK': 'Czech Koruna',
                  'DKK': 'Danish Krone',
                  'EUR': 'Euro',
                  'GBP': 'British Pound',
                  'HKD': 'Hong Kong Dollar',
                  'HRK': 'Croatian Kuna',
                  'HUF': 'Hungarian Forint',
                  'IDR': 'Indonesian Rupiah',
                  'ILS': 'Israeli New Shekel',
                  'INR': 'Indian Rupee',
                  'ISK': 'Icelandic Króna',
                  'JPY': 'Japanese Yen',
                  'KRW': 'South Korean Won',
                  'MXN': 'Mexican Peso',
                  'MYR': 'Malaysian Ringgit',
                  'NOK': 'Norwegian Krone',
                  'NZD': 'New Zealand Dollar',
                  'PHP': 'Philippine Piso',
                  'PLN': 'Polish Zloty',
                  'RON': 'Romanian Leu',
                  'SEK': 'Swedish Krona',
                  'SGD': 'Singapore Dollar',
                  'THB': 'Thai Baht',
                  'TRY': 'Turkish Lira',
                  'USD': 'US Dollar',
                  'ZAR': 'South African Rand',
                  }



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (c) 2020-2021 Salvador E. Tropea
# Copyright (c) 2020-2021 Instituto Nacional de Tecnología Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
Log module

Handles logging initialization and formating.
"""
import io
import os
import sys
import traceback
import logging
no_colorama = False
try:
    from colorama import init as colorama_init, Fore, Back, Style
except ImportError:
    no_colorama = True
# If colorama isn't installed use an ANSI basic replacement
if no_colorama:
    from .ansi import Fore, Back, Style  # noqa: F811
else:
    colorama_init()
# Default domain, base name for the tool
domain = 'kilog'
filters = None

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft/'
__company__ = 'INTI-CMNB - Argentina'


def get_logger(name=None):
    """Get a module for a submodule or the root logger if no name is provided"""
    # print('get_logger '+str(name))
    if name:
        if name.startswith(domain):
            return logging.getLogger(name)
        return logging.getLogger(domain+'.'+name)
    return logging.getLogger(domain)


def set_domain(name):
    """Set the base name for the tool"""
    global domain
    domain = name


def set_filters(f):
    """Set the list of warning filters"""
    global filters
    filters = f


class MyLogger(logging.Logger):
    warn_hash = {}
    warn_tcnt = warn_cnt = n_filtered = 0

    def warning(self, msg, *args, **kwargs):
        MyLogger.warn_tcnt += 1
        # Get the message applying optional C style expansions
        # No longer used:
        # if isinstance(msg, str) and len(args):
        #     buf = StringIO()
        #     buf.write(msg % args)
        #     buf = buf.getvalue()
        # else:
        buf = str(msg)
        # Avoid repeated warnings
        if buf in MyLogger.warn_hash:
            MyLogger.warn_hash[buf] += 1
            return
        # Apply the filters
        if filters and buf.startswith('(WC'):
            pos_end = buf.find(')')
            if pos_end > 0:
                number = int(buf[3:pos_end])
                for f in filters:
                    if f.number == number and f.regex.search(buf):
                        MyLogger.n_filtered += 1
                        return
        MyLogger.warn_cnt += 1
        MyLogger.warn_hash[buf] = 1
        if sys.version_info.major > 3 or (sys.version_info.major == 3 and sys.version_info.minor >= 8):
            super(self.__class__, self).warning(buf, stacklevel=2, **kwargs)  # pragma: no cover (Py38)
        else:
            super(self.__class__, self).warning(buf, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        if level < self.getEffectiveLevel():
            return
        if sys.version_info.major > 3 or (sys.version_info.major == 3 and sys.version_info.minor >= 8):
            super(self.__class__, self).debug(msg, stacklevel=2, *args, **kwargs)  # pragma: no cover (Py38)
        else:
            super(self.__class__, self).debug(msg, *args, **kwargs)

    def log_totals(self):
        if MyLogger.warn_cnt:
            filt_msg = ''
            if MyLogger.n_filtered:
                filt_msg = ', {} filtered'.format(MyLogger.n_filtered)
            self.info('Found {} unique warning/s ({} total{})'.format(MyLogger.warn_cnt, MyLogger.warn_tcnt, filt_msg))

    def findCaller(self, stack_info=False, stacklevel=1):
        f = sys._getframe(1)
        # Skip frames from logging module
        while '/logging/' in os.path.normcase(f.f_code.co_filename):
            f = f.f_back
        # Apply the indicated stacklevel
        while stacklevel > 1:
            f = f.f_back
            stacklevel -= 1
        # Skip the __init__.py wrappers
        fname = os.path.normcase(f.f_code.co_filename)
        if fname.endswith('__init__.py') or fname.endswith('log__.py'):
            f = f.f_back
        # Create the stack info if needed
        sinfo = None
        if stack_info:
            out = io.StringIO()
            out.write(u"Stack (most recent call last):\n")
            traceback.print_stack(f, file=out)
            sinfo = out.getvalue().rstrip(u"\n")
        return os.path.normcase(f.f_code.co_filename), f.f_lineno, f.f_code.co_name, sinfo


def set_verbosity(logger, debug, quiet):
    # Choose the log level
    if debug is not None:
        log_level = logging.DEBUG + 1 - debug
    elif quiet is True:
        log_level = logging.ERROR
    else:
        log_level = logging.INFO
    logger.setLevel(log_level)


class FilterOnlyInfo(object):
    def filter(self, record):
        return record.levelno == logging.INFO


class FilterNoInfo(object):
    def filter(self, record):
        return record.levelno != logging.INFO


def init():
    """Initialize the logging feature using a custom format"""
    # Use a class to count and filter warnings
    logging.setLoggerClass(MyLogger)
    logger = get_logger()
    # Handler for all but info.
    # Outputs to stderr
    ch = logging.StreamHandler()
    ch.addFilter(FilterNoInfo())
    ch.setFormatter(CustomFormatter(sys.stderr))
    logger.addHandler(ch)
    # Handler for t info.
    # Outputs to stdout
    ch = logging.StreamHandler(sys.stdout)
    ch.addFilter(FilterOnlyInfo())
    ch.setFormatter(CustomFormatter(sys.stdout))
    logger.addHandler(ch)
    return logger


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors"""

    def __init__(self, stream):
        super(logging.Formatter, self).__init__()
        if stream.isatty():
            white = Fore.WHITE
            yellow = Fore.YELLOW + Style.BRIGHT
            red = Fore.RED + Style.BRIGHT
            red_alarm = Fore.RED + Back.WHITE + Style.BRIGHT
            cyan = Fore.CYAN + Style.BRIGHT
            reset = Style.RESET_ALL
        else:
            white = ""
            yellow = ""
            red = ""
            red_alarm = ""
            cyan = ""
            reset = ""
        # format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s "
        #          "(%(filename)s:%(lineno)d)"
        format = "%(levelname)s:%(message)s (%(name)s - %(filename)s:%(lineno)d)"
        format_simple = "%(message)s"

        self.FORMATS = {
            logging.DEBUG: cyan + format + reset,
            logging.INFO: white + format_simple + reset,
            logging.WARNING: yellow + format + reset,
            logging.ERROR: red + format + reset,
            logging.CRITICAL: red_alarm + format + reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Salvador E. Tropea
# Copyright (c) 2021 Instituto Nacional de Tecnología Industrial
# License: Apache 2.0
# Project: KiCost
# Adapted from: https://github.com/alexprengere/currencyconverter
"""
CurrencyConverter:
This is reduced version of the 'Currency Converter' by Alex Prengère.
Original project: https://github.com/alexprengere/currencyconverter

This version only supports conversions for the last exchange rates, not
historic ones.

On the other hand this version always tries to get the last rates.
-----------------------------------------------------------------------
list_currencies, get_currency_symbol get_currency_name and
format_currency:

These functions are replacements for Babel
(http://babel.pocoo.org/en/latest/index.html).

Babel is really nice, but a huge overkill for what we need. In
particular KiCost cunrrently supports only 'en_US'.
"""
try:
    from .default_rates import default_rates, default_date
except ImportError:
    # Only useful to boostrap
    default_rates = {}
    default_date = ''
from .download_rates import download_rates
try:
    from .currency_tables import currency_symbols, currency_names
except ImportError:
    # Only useful to boostrap
    currency_symbols = {}
    currency_names = {}

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft/'
__company__ = 'INTI-CMNB - Argentina'


class CurrencyConverter(object):
    def __init__(self):
        self.initialized = False

    def _do_init(self):
        if self.initialized:
            return
        self.date, self.rates = download_rates()
        if not self.date:
            self.date = default_date
            self.rates = default_rates
        self.initialized = True

    def convert(self, amount, currency, new_currency='EUR'):
        """Convert amount from a currency to another one.

        :param float amount: The amount of `currency` to convert.
        :param str currency: The currency to convert from.
        :param str new_currency: The currency to convert to.

        :return: The value of `amount` in `new_currency`.
        :rtype: float

        >>> c = CurrencyConverter()
        >>> c.convert(100, 'EUR', 'USD')
        """
        self._do_init()
        for c in currency, new_currency:
            if c not in self.rates:
                raise ValueError('{0} is not a supported currency'.format(c))

        r0 = self.rates[currency]
        r1 = self.rates[new_currency]

        return float(amount) / r0 * r1


def list_currencies():
    ''' Get a list of known currencies '''
    return currency_symbols.keys()


def get_currency_symbol(currency, locale=None):
    ''' Get the symbol to represent the specified ISO currency '''
    return currency_symbols.get(currency, '$')


def get_currency_name(currency, locale=None):
    ''' Get the name for the specified ISO currency '''
    return currency_names.get(currency, '')


def format_currency(price, currency, locale=None):
    ''' Format price for the specified ISO currency '''
    return currency_symbols.get(currency, '$') + '{:,.2f}'.format(price)



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
  @package
  KiCost script, also with Graphical User Interface - GUI, under MIT license for generate part-cost spreadsheets for circuit boards developed with KiCad and
  others EDAs.
  Full manual at https://xesscorp.github.io/KiCost
  Development at https://github.com/xesscorp/KiCost
  KiCost is powered by the Kitspace PartInfo API (https://kitspace.org/). Partinfo hooks into paid-for 3rd party services. If you find KiCost useful please
  donate to the Kitspace Open Collective to keep the service running. (https://opencollective.com/kitspace).

  Command line:
      kicost -i "%I" "%O.xslx"
  Run Graphical User Interface:
      kicost
  Check for help on terminal:
      kicost --help
"""


# Libraries.
import sys
import os
import pprint
from collections import OrderedDict
from copy import copy
from math import ceil

# Stops UnicodeDecodeError exceptions.
try:
    reload(sys)
    sys.setdefaultencoding('utf8')
except NameError:
    pass  # Happens if reload is attempted in Python 3.

# Only export this routine for use by the outside world.
__all__ = ['kicost', 'output_filename', 'kicost_gui_notdependences', 'query_part_info']

from .global_vars import DEFAULT_CURRENCY, SEPRTR, DEF_MAX_COLUMN_W, ERR_KICOSTCONFIG, ERR_ARGS, W_TRANS, W_NOMANP
# * Import the KiCost libraries functions.
# Import information for various EDA tools.
from .edas.tools import field_name_translations, subpartqty_split, group_parts, PRJ_STR_DECLARE, PRJPART_SPRTR, partgroup_qty
from .edas import get_part_groups
# Creation of the final XLSX spreadsheet.
from .spreadsheet import create_spreadsheet, Spreadsheet
# Import the scrape API
from .distributors import get_dist_parts_info, get_registered_apis, get_distributors_iter, get_distributor_info
from . import debug_detailed, debug_overview, is_debug_detailed, is_debug_overview, error, warning, KiCostError


def query_part_info(parts, dist_list, currency=DEFAULT_CURRENCY):
    """ Used to get the parts price and information.
        Also used from outside KiCost. """
    if is_debug_overview():
        api_list = [d.name + ('(Disabled)' if not d.enabled else '') for d in get_registered_apis()]
        debug_overview('Scrape API list ' + str(api_list))
    get_dist_parts_info(parts, dist_list, currency)


def solve_parts_qtys(parts, multi_prj, prj_info):
    """ Used to fill the qty_str, qty and qty_total_spreadsheet members of the parts.
        Also used from outside KiCost. """
    for part in parts:
        part.qty_str, part.qty = partgroup_qty(part)
        # Total for all boards
        if multi_prj:
            total = 0
            for i_prj, p_info in enumerate(prj_info):
                total += part.qty[i_prj] * p_info['qty']
        else:
            total = part.qty * prj_info[0]['qty']
        part.qty_total_spreadsheet = ceil(total)


def kicost(in_file, eda_name, out_filename, user_fields, ignore_fields, group_fields, translate_fields, variant, dist_list, collapse_refs=True,
           suppress_cat_url=True, currency=DEFAULT_CURRENCY, max_column_width=DEF_MAX_COLUMN_W, split_extra_fields=[],
           board_qty=[Spreadsheet.DEFAULT_BUILD_QTY]):
    ''' @brief Run KiCost.

    Take a schematic input file and create an output file with a cost spreadsheet in xlsx format.

    @param in_file `list(str())` List of the names of the input BOM files.
    @param eda_name `list(str())` of the EDA modules to be used to open the `in_file`list.
    @param out_filename `str()` XLSX output file name.
    @param user_fields `list()` of the user fields to be included on the spreadsheet global part.
    @param ignore_fields `list()` of the fields to be ignored on the read EDA modules.
    @param group_fields `list()` of the fields to be grouped/merged on the function group parts that
    are not grouped by default.
    @param translate_fields `list()` of the fields to translate to translate or remove (if `~` present).
    @param variant `list(str())` of regular expression to the BOM variant of each file in `in_file`.
    @param dist_list `list(str())` to be scraped
    modules. If `None`, no web/local distributors will be scraped.
    @param collapse_refs `bool()` Collapse or not the designator references in the spreadsheet.
    Default `True`.
    @param suppress_cat_url `bool()` Suppress the distributors catalogue links into the catalogue code in the spreadsheet.
    Default `True`.
    @param currency `str()` Currency in ISO4217. Default 'USD'.
    @param max_column_width `int()` The maximum column width. If 0 disables cell size adjust. Default: DEF_MAX_COLUMN_W
    @param split_extra_fields `list(str())` Fields that will be split using the multipart mechanism.
    @param board_qty `list(int())` Board quantities for each project.
    '''
    # Add or remove field translations, ignore in case the trying to
    # re-translate default field names.
    if translate_fields:
        if len(translate_fields) % 2 == 1:
            raise KiCostError('Translation fields argument should have an even number of words.', ERR_ARGS)
        for c in range(0, len(translate_fields), 2):
            # field_name_translations.keys(), field_name_translations.values()
            if translate_fields[c] in field_name_translations.values():
                warning(W_TRANS, "Unable to re-translate \"{}\" to \"{}\", this is used as an internal field name.".
                        format(translate_fields[c].lower(), translate_fields[c+1].lower()))
                continue
            if translate_fields[c+1] != '~':
                field_name_translations.update({translate_fields[c].lower(): translate_fields[c+1].lower()})
            else:
                field_name_translations.pop(translate_fields[c].lower(), None)

    # Check the integrity of the user personal fields, this should not
    # be any of the reserved fields.
    # This is checked after the translation `dict` is complete, so an
    # before used name field on the translate dictionary can be used
    # user field.
    user_fields = list(OrderedDict([(lv, 1) for lv in user_fields]))  # Avoid repeated fields
    for f in user_fields:
        if f.lower() in field_name_translations:
            warning(W_TRANS, "\"{f}\" field is a reserved field and can not be used as user field."
                    " Try to remove it from internal dictionary using `--translate_fields {f} ~`".format(f=f.lower()))
            user_fields.remove(f)

    c_files = len(in_file)
    # Deal with some code exception (only one EDA tool or variant
    # informed in the multiple BOM files input).
    # This code is never used when called from __main__
    if not isinstance(in_file, list):
        in_file = [in_file]
    if not isinstance(variant, list):
        variant = [variant] * c_files
    elif len(variant) != c_files:
        variant = [variant[0]] * c_files  # Assume the first as default.
    if not isinstance(eda_name, list):
        eda_name = [eda_name] * c_files
    elif len(eda_name) != c_files:
        eda_name = [eda_name[0]] * c_files  # Assume the first as default.
    if not isinstance(board_qty, list):
        board_qty = [board_qty] * c_files
    elif len(board_qty) != c_files:
        board_qty = [board_qty[0]] * c_files  # Assume the first as default.

    # Get groups of identical parts.
    parts = OrderedDict()
    prj_info = list()
    for i_prj in range(c_files):
        p, info = get_part_groups(eda_name[i_prj], in_file[i_prj], ignore_fields, variant[i_prj], dist_list)
        p = subpartqty_split(p, dist_list, split_extra_fields)
        if c_files > 1:
            # In the case of multiple BOM files, add the project prefix identifier
            # to each reference/designator. Use the field 'manf#_qty' to set
            # the quantity for each project, creating a `list()` with length
            # of number of BOM files. This vector will be used in the `group_parts()`
            # to create groups with elements of same 'manf#' that came from different
            # projects.
            debug_overview('Multi BOMs detected, attaching project identification to references...')
            qty_base = ['0'] * c_files  # Zero for all projects
            for p_ref, p_fields in p.items():
                # Copy the qty to the position corresponding to this project
                qty_base[i_prj] = p_fields.get('manf#_qty', '1')
                p_fields['manf#_qty'] = copy(qty_base)
                parts[PRJ_STR_DECLARE + str(i_prj) + PRJPART_SPRTR + p_ref] = p_fields
        else:
            parts.update(p)
        info['qty'] = board_qty[i_prj]
        prj_info.append(info)

    # Group part out of the module to be possible to merge different
    # project lists, ignore some field to merge given in the `group_fields`.
    FIELDS_SPREADSHEET = ['refs', 'value', 'desc', 'footprint', 'manf', 'manf#']
    FIELDS_MANFCAT = ([d + '#' for d in get_distributors_iter()] + ['manf#'])
    FIELDS_MANFQTY = ([d + '#_qty' for d in get_distributors_iter()] + ['manf#_qty'])
    FIELDS_IGNORE = FIELDS_SPREADSHEET + FIELDS_MANFCAT + FIELDS_MANFQTY + user_fields + ['pricing']
    for ref, fields in list(parts.items()):
        for f in fields:
            # Merge all extra fields that read on the files that will
            # not be displayed (Needed to check `user_fields`).
            if f not in FIELDS_IGNORE and SEPRTR not in f and f not in group_fields:
                # Not include repetitive field names or fields with the separator `:` defined on `SEPRTR`.
                group_fields += [f]

    # Some fields to be merged on specific EDA are enrolled bellow.
    if 'kicad' in eda_name:
        group_fields += ['libpart']  # This field may be a mess on multiple sheet designs.
    if len(set(eda_name)) > 2:
        # If more than one EDA software was used, ignore the 'footprint'
        # field, because they could have different libraries names.
        group_fields += ['footprint']
    # Always ignore 'desc' ('description') and 'var' ('variant') fields, merging the components in groups.
    group_fields += ['desc', 'var']
    group_fields = set(group_fields)
    parts = group_parts(parts, group_fields, c_files)

    # If do not have the manufacture code 'manf#' and just distributors codes,
    # check if is asked to scrape a distributor that do not have any code in the
    # parts so, exclude this distributors for the scrap list. This decrease the
    # warning messages given during the process.
    all_fields = set()
    for p in parts:
        all_fields.update(p.fields)
    if 'manf#' not in all_fields:
        new_list = []
        for d in dist_list:
            if d+'#' not in all_fields:
                warning(W_NOMANP, "No 'manf#' and '{}#' field in any part: no information by '{}'.".
                        format(d, get_distributor_info(d).label.name))
            else:
                new_list.append(d)
        dist_list = new_list
    # Debug the resulting list
    if is_debug_detailed():
        debug_detailed('Distributors: ' + pprint.pformat(dist_list))
    #
    # Solve the quantity for each group
    #
    # Make sure we have the number of boards for each project
    for p_info in prj_info:
        if 'qty' not in p_info:
            p_info['qty'] = Spreadsheet.DEFAULT_BUILD_QTY
    multi_prj = len(prj_info) > 1
    # Compute the qtys
    solve_parts_qtys(parts, multi_prj, prj_info)
    # Get the distributor pricing/etc for each part.
    query_part_info(parts, dist_list, currency)

    # Create the part pricing spreadsheet.
    create_spreadsheet(parts, prj_info, out_filename, dist_list, currency, collapse_refs, suppress_cat_url,
                       user_fields, '-'.join(variant) if len(variant) > 1 else variant[0], max_column_width)

    # Print component groups for debugging purposes.
    if is_debug_detailed():
        for part in parts:
            for f in dir(part):
                if f.startswith('__'):
                    continue
                elif f.startswith('html_trees'):
                    continue
                else:
                    head = '{} = '.format(f)
                    try:
                        debug_detailed(head + pprint.pformat(part.__dict__[f]))
                    except TypeError:
                        # Python 2.7 pprint has some problem ordering None and strings.
                        debug_detailed(head + str(part.__dict__[f]))
                    except KeyError:
                        pass
            debug_detailed('')


# Maximum length of the name of the spreadsheet output generate, this is used in the multifiles to limit the
# automatic name generation.
FILE_OUTPUT_MAX_NAME = 10
# Minimum length of characters to use of the input files to create the name of the spreadsheet output file.
# This is used in the multifile BoM and have prioritize in the `FILE_OUTPUT_MAX_NAME` definition.
FILE_OUTPUT_MIN_INPUT = 5
# Separator in the name of the output spreadsheet file when used multiple input file to generate automatically
# the name.
FILE_OUTPUT_INPUT_SEP = '-'


# Here because is used at `__main__.py` and `kicost_gui.py`.
def output_filename(files_input):
    ''' @brief Compose a name with the multiple BOM input file names.
    Compose a name with the multiple BOM input file names, limiting to,
    at least, the first `FILE_OUTPUT_MIN_INPUT` characters of each name
    (avoid huge names by `FILE_OUTPUT_MAX_NAME`definition). Join the names
    of the input files by `FILE_OUTPUT_INPUT_SEP` definition.
    The output folder is the folder of the first file.
    @param files_input `list()`of the input file names.
    @return `str()` file name for the spreadsheet.
    '''

    if len(files_input) == 1:
        # Use the folder of the project.
        return os.path.splitext(files_input[0])[0] + '.xlsx'
    else:
        # If more the one file selected, check if they are in
        # the same folder, if don't, output in the folder where
        # `kicost` was called.
        dir_output = os.path.dirname(files_input[0]) + os.path.sep
        for dir_idx in range(len(files_input)):
            if os.path.dirname(files_input[dir_idx]) != dir_output:
                dir_output = os.getcwd()

    _end = max(int(FILE_OUTPUT_MAX_NAME/len(files_input)), FILE_OUTPUT_MIN_INPUT-len(FILE_OUTPUT_INPUT_SEP))
    file_name = FILE_OUTPUT_INPUT_SEP.join([os.path.splitext(os.path.basename(input_name))[0][:_end] for input_name in files_input])
    file_output = os.path.join(dir_output, file_name + '.xlsx')
    return file_output


def kicost_gui_notdependences():
    error('You don\'t have the wxPython dependence to run the GUI interface.')
    error('Run once of the following commands in a terminal to install it:')
    error('pip3 install -U wxPython # For Windows & macOS')
    error('sudo apt-get install python3-wxgtk4.0 # Modern Linux derived from Debian, like Ubuntu')
    error('pip install -U -f https://extras.wxpython.org/wxPython4/extras/linux/gtk3/ubuntu-16.04 wxPython # For Linux 16.04')
    error('Or download the last version from <https://wxpython.org/pages/downloads/>')
    sys.exit(ERR_KICOSTCONFIG)



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Junior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This isn't currently used
# # Extra informations to by got by each part in the distributors.
# EXTRA_INFO_DIST = ['value', 'tolerance', 'footprint', 'power', 'current', 'voltage', 'frequency', 'temp_coeff', 'manf',
#                    'size', 'op temp', 'orientation', 'color',
#                    'datasheet', 'image',  # Links.
#                    ]
# extra_info_dist_name_translations = {
#     # TODO it will need to put here language translation after implementation of ISSUE #65?
#     'resistance': 'value',
#     'inductance': 'value',
#     'capacitance': 'value',
#     'manufacturer': 'manf',
#     'package': 'footprint',
#     'package / case': 'footprint',
#     'datasheets': 'datasheet',
#     'dimension': 'size',
#     'size / dimension': 'size',
#     'operating temperature': 'op temp',
#     'voltage - rated': 'voltage',
#     'Mating Orientation': 'orientation',
#     'coulor': 'color',
#     'wire gauge': 'wire',
# }



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2022 by Salvador E. Tropea / Instituto Nacional de Tecnologia Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Simple wrappers for the log functions in the main module

import logging

DEBUG_OVERVIEW = logging.DEBUG
DEBUG_DETAILED = logging.DEBUG-1
DEBUG_OBSESSIVE = logging.DEBUG-2
DEBUG_HTTP_HEADERS = logging.DEBUG-3
DEBUG_HTTP_RESPONSES = logging.DEBUG-4
DEBUG_FULL = logging.DEBUG-9
# Minimum possible log level is logging.DEBUG-9 !


# The root logger of the application. This has to be the root logger to catch
# output from libraries (e.g. requests) as well.
main_logger = None


def debug(*args):
    main_logger.log(*args)


def debug_detailed(*args):
    main_logger.log(DEBUG_DETAILED, *args)


def is_debug_detailed():
    return main_logger.getEffectiveLevel() <= DEBUG_DETAILED


def debug_overview(*args):
    main_logger.log(DEBUG_OVERVIEW, *args)


def is_debug_overview():
    return main_logger.getEffectiveLevel() <= DEBUG_OVERVIEW


def debug_obsessive(*args):
    main_logger.log(DEBUG_OBSESSIVE, *args)


def is_debug_obsessive():
    return main_logger.getEffectiveLevel() <= DEBUG_OBSESSIVE


def debug_full(*args):
    main_logger.log(DEBUG_FULL, *args)


def debug_general(*args):
    main_logger.debug(*args)


def info(*args):
    main_logger.info(*args)


def warning(code, msg):
    main_logger.warning(code + msg)


def error(*args):
    main_logger.error(*args)


def get_logger():
    global main_logger
    return main_logger


def set_logger(logger):
    global main_logger
    main_logger = logger



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo G Jr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from .eda import eda_class
# Import and register here the file read modules.
from .eda_kicad import eda_kicad  # noqa: F401
from .eda_altium import eda_altium  # noqa: F401
from .generic_csv import generic_csv  # noqa: F401

# Author information.
__author__ = 'Hildo Guillardi Junior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'


def get_registered_eda_names():
    ''' Get the names of the registered EDAs '''
    return eda_class.registered.keys()


def get_registered_eda_labels():
    ''' Get the labels of the registered EDAs '''
    return [eda.label for eda in eda_class.registered.values()]


def get_part_groups(eda, in_file, ignore_fields, variant, distributors):
    ''' Get the parts for a file using the indicated EDA '''
    return eda_class.get_part_groups(eda, in_file, ignore_fields, variant, distributors)


def file_eda_match(file_name):
    ''' Check which EDA is suitable for this file '''
    return eda_class.file_eda_match(file_name)


def get_eda_label(name):
    ''' Returns a beautiful name for the EDA '''
    return eda_class.registered[name].label


def set_edas_logger(logger):
    ''' Sets the logger used by the class '''
    eda_class.logger = logger



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Júnior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'
# This script aims to be a Eeschema configuration layer.

# Python libraries.
import os
import sys
import re

try:
    import sexpdata  # Try to use a external updated library.
except ImportError:
    from . import sexpdata  # Use the local file.
from .global_vars import PLATFORM_MACOS_STARTS_WITH, PLATFORM_WINDOWS_STARTS_WITH, ERR_KICADCONFIG  # Debug, language and default configurations.
from . import KiCostError

__all__ = ['get_app_config_path',
           'PATH_KICAD_CONFIG', 'PATH_EESCHEMA_CONFIG',
           'bom_plugin_add_entry', 'bom_plugin_remove_entry',
           'fields_add_entry', 'fields_remove_entry']


def get_app_config_path(appname):
    if sys.platform == PLATFORM_MACOS_STARTS_WITH:
        from AppKit import NSSearchPathForDirectoriesInDomains
        # http://developer.apple.com/DOCUMENTATION/Cocoa/Reference/Foundation/Miscellaneous/Foundation_Functions/Reference/reference.html#//apple_ref/c/func/NSSearchPathForDirectoriesInDomains
        # NSApplicationSupportDirectory = 14
        # NSUserDomainMask = 1
        # True for expanding the tilde into a fully qualified path
        appdata = os.path.join(NSSearchPathForDirectoriesInDomains(5, 1, True)[0], "Preferences", appname)
    elif sys.platform == PLATFORM_WINDOWS_STARTS_WITH:
        appdata = os.path.join(os.environ['APPDATA'], appname)
    else:
        # ~/.config/kicad
        appdata = os.path.expanduser(os.path.join("~", ".config", appname))
    return appdata


PATH_KICAD_CONFIG = get_app_config_path('kicad')
if not PATH_KICAD_CONFIG:
    raise KiCostError('KiCad configuration folder not found.', ERR_KICADCONFIG)
PATH_EESCHEMA_CONFIG = os.path.join(PATH_KICAD_CONFIG, "eeschema")


def get_user_documents():
    if sys.platform == PLATFORM_MACOS_STARTS_WITH:
        user_documents = os.path.expanduser(os.path.join("~", "Documents"))
    elif sys.platform == PLATFORM_WINDOWS_STARTS_WITH:
        # e.g. c:\users\bob\Documents
        user_documents = os.path.join(os.environ['USERPROFILE'], "Documents")
    else:
        user_documents = os.path.expanduser(os.path.join("~", "Documents"))
    return user_documents


# def get_running_processes(appname):
#     processes = []
#     for p in psutil.process_iter():
#         try:
#             if p.name().lower().startswith(appname):
#                 processes.append(p)
#         except psutil.Error:
#             pass
#     return processes


def before(value, a):
    # Find first part and return slice before it.
    pos_a = value.find(a)
    if pos_a == -1:
        return ""
    return value[0:pos_a]


def after(value, a):
    # Find and validate first part.
    pos_a = value.find(a)
    if pos_a == -1:
        return ""
    # Returns chars after the found string.
    adjusted_pos_a = pos_a + len(a)
    if adjusted_pos_a >= len(value):
        return ""
    return value[adjusted_pos_a:]


def de_escape(s):
    result = ""
    in_escape = False
    for c in s:
        if in_escape:
            result += c
            in_escape = False
        else:
            if c == '\\':
                in_escape = True
            else:
                result += c
    return result


def escape(s):
    result = ""
    for c in s:
        result += '\\' + c if c == '\\' else c
    return result


def get_config_item(config, key):
    for p in config:
        if before(p, '=').strip() == key:
            return after(p, '=')


def update_config_file(config, key, value):
    new_config = []
    for p in config:
        if before(p, '=') == key:
            new_config.append(key + '=' + value)
        else:
            new_config.append(p)
    return new_config


def write_config_file(path, config):
    with open(path, "w") as f:
        f.write('\n'.join(config))


def read_config_file(path):
    with open(path) as f:
        config = f.read().split('\n')
    return config


def fields_add_entry(values_modify, re_flags=re.IGNORECASE):
    '''Add a list of fields to the Eeschema template.'''
    if type(values_modify) is not list:
        values_modify = [values_modify]
    config = read_config_file(PATH_EESCHEMA_CONFIG)
    values = [p for p in config if p.startswith("FieldNames")]
    changes = False
    if len(values) == 1:
        values = after(values[0], "FieldNames=")
        values = de_escape(values)
        values = sexpdata.loads(values)
        # TODO validate it, was using an undefined variable `name`
        if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
            values = values.replace("\\", '/')
        for idx, value_modify in enumerate(values_modify):
            value_found = False
            for idx, value in enumerate(values[1:]):
                search = value[1]
                if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
                    search = value[1].replace("\\", '/')
                if re.findall(value_modify, search[1], re_flags):
                    value_found = True
            if not value_found:
                values.append([sexpdata.Symbol('field'), [sexpdata.Symbol('name'), value_modify]])
                changes = True
    if changes:
        s = sexpdata.dumps(values)
        config = update_config_file(config, "FieldNames", escape(s))
    write_config_file(PATH_EESCHEMA_CONFIG, config)


def fields_remove_entry(values_modify, re_flags=re.IGNORECASE):
    '''Remove a list of fields from the Eeschema template.'''
    if type(values_modify) is not list:
        values_modify = [values_modify]
    config = read_config_file(PATH_EESCHEMA_CONFIG)
    values = [p for p in config if p.startswith("FieldNames")]
    changes = False
    delete_list = []
    if len(values) == 1:
        values = after(values[0], "FieldNames=")
        values = de_escape(values)
        values = sexpdata.loads(values)
        # TODO validate it, was using an undefined variable `name`
        if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
            values = values.replace("\\", '/')
        for value_modify in values_modify:
            for idx, value in enumerate(values[1:]):
                search = value[1]
                if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
                    search = value[1].replace("\\", '/')
                if re.findall(value_modify, search[1], re_flags):
                    changes = True  # The name in really in the 'name'.
                    delete_list.append(idx)  # We want to delete this entry.
        for delete in sorted(set(delete_list), reverse=True):
            del values[delete+1]
    if changes:
        s = sexpdata.dumps(values)
        config = update_config_file(config, "FieldNames", escape(s))
    write_config_file(PATH_EESCHEMA_CONFIG, config)


def bom_plugin_remove_entry(name, re_flags=re.IGNORECASE):
    '''Remove a BOM plugin entry to the Eeschema configuration file.'''
    config = read_config_file(PATH_EESCHEMA_CONFIG)
    bom_plugins_raw = [p for p in config if p.startswith("bom_plugins")]
    new_list = []
    new_list.append(sexpdata.Symbol("plugins"))
    changes = False
    if len(bom_plugins_raw) == 1:
        bom_plugins_raw = after(bom_plugins_raw[0], "bom_plugins=")
        bom_plugins_raw = de_escape(bom_plugins_raw)
        bom_list = sexpdata.loads(bom_plugins_raw)
        if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
            name = name.replace("\\", '/')
        for plugin in bom_list[1:]:
            search = plugin[1]
            if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
                search = plugin[1].replace("\\", '/')
            if re.findall(name, search, re_flags):
                changes = True  # The name in really in the 'name'.
                continue  # We want to delete this entry.
            else:
                for entry in plugin[2:]:
                    if entry[0] == sexpdata.Symbol('opts') and re.findall(r'nickname\s*=\s*'+name, entry[1], re_flags):
                        changes = True
                        continue  # The name is in the 'nickname'.
                new_list.append(plugin)  # This plugin remains on the list.
    if changes:
        s = sexpdata.dumps(new_list)
        config = update_config_file(config, "bom_plugins", escape(s))
    write_config_file(PATH_EESCHEMA_CONFIG, config)


def bom_plugin_add_entry(name, cmd, nickname=None, re_flags=re.IGNORECASE, put_first=True, set_default=True):
    '''Add a BOM plugin entry to the Eeschema configuration file.'''
    config = read_config_file(PATH_EESCHEMA_CONFIG)
    bom_plugins_raw = [p for p in config if p.startswith("bom_plugins")]
    new_list = []
    new_list.append(sexpdata.Symbol("plugins"))
    if len(bom_plugins_raw) == 1:
        bom_plugins_raw = after(bom_plugins_raw[0], "bom_plugins=")
        bom_plugins_raw = de_escape(bom_plugins_raw)
        bom_list = sexpdata.loads(bom_plugins_raw)
        if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
            name = name.replace("\\", '/')
        for plugin in bom_list[1:]:
            search = plugin[1]
            if sys.platform.startswith(PLATFORM_WINDOWS_STARTS_WITH):
                search = plugin[1].replace("\\", '/')
            if re.findall(name, search, re_flags):
                if not nickname:
                    return  # Plugin already added and don't have nickname.
                for entry in plugin[2:]:
                    if entry[0] == sexpdata.Symbol('opts') and re.findall(r'nickname\s*=\s*'+nickname, entry[1], re_flags):
                        return  # Plugin already added with this nickname.
            new_list.append(plugin)
    if not nickname:
        new_list.append([sexpdata.Symbol('plugin'), sexpdata.Symbol(name), [sexpdata.Symbol('cmd'), cmd]])
    else:
        new_list.append([sexpdata.Symbol('plugin'), name,
                        [sexpdata.Symbol('cmd'), cmd],
                        [sexpdata.Symbol('opts'), 'nickname={}'.format(nickname)]])
    if len(new_list):
        # Put KiCost at first.
        if put_first:
            new_list.insert(1, new_list[-1])
            del new_list[-1]
    config = update_config_file(config, "bom_plugins", escape(sexpdata.dumps(new_list)))
    write_config_file(PATH_EESCHEMA_CONFIG, config)
    if set_default:
        import fileinput
        for line in fileinput.input(PATH_EESCHEMA_CONFIG, inplace=True):
            if line.strip().startswith('bom_plugin_selected='):
                line = 'bom_plugin_selected={}\n'.format(nickname)
            sys.stdout.write(line)



# -*- coding: utf-8 -*-
# This file keep all the web distributors information used by the different API/scrap modules.

# Author information.
__author__ = 'Hildo Guillardi Júnior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'

# Used as identification for all user fields allowed for some custom importation in some distributors.
# It is used a low probability "word" corresponding to all user fields.
ORDER_COL_USERFIELDS = '*__USER__FIELDS__*'
# TODO at the GUI, could be a tab with to personalize this configuration, using this file as default, and the user could include or exclude some
#      personal field.


class DistributorOrder(object):
    '''@brief Class to indicate how to place an order in a distributor.'''
    def __init__(self, url=None, cols=[], header=None, delimiter=',', replace_by_char='; ', not_allowed_char=',\n', info=None, limit=None):
        self.url = url
        # Sort-order fields for online orders. The not present fields are by-passed and `None` represent a empty column.
        self.cols = cols
        # Header to help user undertanding (used in some importations). **Currently unused**
        self.header = header
        # Delimiter for online orders.
        self.delimiter = delimiter
        # The `delimiter` is not allowed inside description. This character is used to replace it.
        self.replace_by_char = replace_by_char
        # Characters not allowed at the BoM for web-site import.
        self.not_allowed_char = not_allowed_char
        # Descriptive fields size limit
        self.limit = limit
        self.info = info


class DistributorLabel(object):
    '''@brief Class to describe a distributor column.'''
    def __init__(self, name, url, bg, fg='white'):
        self.name = name
        self.url = url
        # Formatting for distributor header in worksheet; bold, font and align are
        # `spreadsheet.py` defined but can by overload here.
        self.format = {'font_color': fg, 'bg_color': bg}


class DistributorInfo(object):
    '''@brief Class to describe a distributor.'''
    def __init__(self, order, label, type='web', ignore_cat=None):
        self.type = type   # Allowable values: 'local' or 'web'.
        self.order = order
        self.label = label
        # Regex used to ignore some catalogue/stock code format.
        # In the Digikey distributor it is used to ignore the Digi-reel package.
        self.ignore_cat = ignore_cat

    def is_local(self):
        return self.type == 'local'

    def is_web(self):
        return self.type == 'web'


distributors_info = {
    'arrow': DistributorInfo(
                order=DistributorOrder(
                    url='https://www.arrow.com/en/bom-tool/',
                    # header='Stock#,Quantity,Designators',
                    cols=['part_num', 'purch', 'refs']),
                label=DistributorLabel('Arrow', 'https://www.arrow.com/', '#000000')),  # Arrow black.
    'digikey': DistributorInfo(
                order=DistributorOrder(
                    url='https://www.digikey.com/ordering/shoppingcart',
                    # header='Quantity,Stock#,Designators',
                    cols=['purch', 'part_num', 'refs']),
                ignore_cat=r'.+(DKR\-ND|\-6\-ND)$',
                label=DistributorLabel('Digi-Key', 'https://www.digikey.com/', '#CC0000')),  # Digi-Key red.
    'farnell': DistributorInfo(
                order=DistributorOrder(
                    url='https://fr.farnell.com/en-FR/quick-order?isQuickPaste=true',
                    # header='Stock#,Quantity,Descriptions,Designators,',
                    cols=['part_num', 'purch', 'desc', 'refs'],
                    limit=30),
                label=DistributorLabel('Farnell', 'https://www.farnell.com/', '#FF6600')),  # Farnell/E14 orange.
    'mouser': DistributorInfo(
                order=DistributorOrder(
                    url='https://mouser.com/bom/',
                    # header='Stock#|Quantity|Designators',
                    cols=['part_num', 'purch', 'refs'],
                    delimiter='|',
                    not_allowed_char='| \n',
                    replace_by_char=';__'),
                label=DistributorLabel('Mouser', 'https://www.mouser.com/', '#004A85')),  # Mouser blue.
    'newark': DistributorInfo(
                order=DistributorOrder(
                    url='https://www.newark.com/quick-order?isQuickPaste=true',
                    # header='Stock#,Quantity,Designators,Descriptions,User',
                    cols=['part_num', 'purch', 'refs', 'desc']),
                label=DistributorLabel('Newark', 'https://www.newark.com/', '#A2AE06')),  # Newark/E14 olive green.
    'rs': DistributorInfo(
                order=DistributorOrder(
                    url='https://uk.rs-online.com/web/mylists/manualQuotes.html?method=showEnquiryCreationPage&mode=new',
                    # header='Stock#,Quantity,-,-,-,Part,Designators',
                    cols=['part_num', 'purch', None, None, None, 'manf#', 'refs']),  # `None` is used for generate a empty column.
                label=DistributorLabel('RS Components', 'https://uk.rs-online.com/', '#FF0000')),  # RS Components red.
    'tme': DistributorInfo(
                order=DistributorOrder(
                    url='https://www.tme.eu/en/Profile/QuickBuy/load.html',
                    # header='Stock# Quantity Designators',
                    cols=['part_num', 'purch', 'refs'],
                    delimiter=' ',
                    not_allowed_char=' \n',
                    replace_by_char=';'),
                label=DistributorLabel('TME', 'https://www.tme.eu/', '#0C4DA1')),  # TME blue.
    'lcsc': DistributorInfo(
                order=DistributorOrder(
                    url='https://lcsc.com/bom.html',
                    header='Quantity,Comment,Designator,Footprint,LCSC Part #(optional)',
                    cols=['purch', 'refs', 'footprint', 'part_num'],
                    info='Copy this header and order to a CSV\n'
                         'file and use it for JLCPCB \n'
                         'manufacturer PCB house.\n'
                         'The multipart components that use\n'
                         '"#" symbol are not allowed by JLCPCB.'),
                label=DistributorLabel('LCSC', 'https://lcsc.com/', '#1166DD')),  # LCSC blue.
    'local_template': DistributorInfo(
                type='local',
                order=DistributorOrder(cols=['part_num', 'purch', 'refs']),
                label=DistributorLabel('Local', None, '#008000')),  # Darker green.
}



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Salvador E. Tropea
# Copyright (c) 2021 Instituto Nacional de Tecnología Industrial
# License: Apache 2.0
# Project: KiCost
"""
Simple helper to download the exchange rates.
"""
import os
import sys
from bs4 import BeautifulSoup

if sys.version_info[0] < 3:
    from urllib2 import urlopen, URLError
else:
    from urllib.request import urlopen, URLError

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft/'
__company__ = 'INTI-CMNB - Argentina'

url = 'https://www.ecb.europa.eu/stats/eurofxref/eurofxref-daily.xml'


def download_rates():
    content = ''
    if os.environ.get('KICOST_CURRENCY_RATES'):
        with open(os.environ['KICOST_CURRENCY_RATES'], 'rt') as f:
            content = f.read()
    else:
        try:
            content = urlopen(url).read().decode('utf8')
        except URLError:
            pass
    soup = BeautifulSoup(content, 'xml')
    rates = {'EUR': 1.0}
    date = ''
    for entry in soup.find_all('Cube'):
        currency = entry.get('currency')
        if currency is not None:
            rates[currency] = float(entry.get('rate'))
        elif entry.get('time'):
            date = entry.get('time')
    return date, rates



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (c) 2021 Salvador E. Tropea
# Copyright (c) 2021 Instituto Nacional de Tecnologïa Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
KiCost test module

Tests for `kicost`. From the root of the projectr run:

pytest-3 --log-cli-level debug
"""

import unittest
import subprocess
import logging
import os
import re
import sys
import shutil
import xml.etree.ElementTree as ET
from kicost.global_vars import ERR_FIELDS

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft/'
__company__ = 'INTI-CMNB - Argentina'

# Collect real world queries (see README.md)
# Change to 1 when the query result must be saved, then revert to 0
ADD_QUERY_TO_KNOWN = 0
# Used to regenerate the references
CREATE_REF = 0
TESTDIR = os.path.dirname(os.path.realpath(__file__))
last_err = None
# Text we want to filter in the XLSX to TXT conversion
XLSX_FILTERS = (('$ date:', None), ('Prj date:', '(file'), ('KiCost', 0))
OCTOPART_KEY = 'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'


def to_str(s):
    if s is None or sys.version_info[0] >= 3:
        return s
    return s.encode('utf-8')


def log_running(what, cmd):
    logging.debug('{} using: {}'.format(what, ' '.join(cmd)))


def xlsx_to_txt(filename, subdir='result_test', sheet=1):
    filename = os.path.join(TESTDIR, filename + '.xlsx')
    logging.debug('Converting to TXT')
    tmpdir = TESTDIR + '/desc'
    assert not os.path.isdir(tmpdir), "Destination for XLSX uncompress is there, remove it and investigate"
    subprocess.call(['unzip', filename, '-d', tmpdir])
    # Some XMLs are stored with 0600
    subprocess.call(['chmod', '-R', 'og+r', tmpdir])
    # Read the table
    worksheet = os.path.join(tmpdir, 'xl', 'worksheets', 'sheet'+str(sheet)+'.xml')
    if not os.path.isfile(worksheet):
        return False
    rows = []
    forms = {}
    root = ET.parse(worksheet).getroot()
    ns = '{http://schemas.openxmlformats.org/spreadsheetml/2006/main}'
    for r in root.iter(ns+'row'):
        rcur = int(r.attrib['r'])
        cols = []
        for cell in r.iter(ns+'c'):
            if 't' in cell.attrib:
                type = cell.attrib['t']
            else:
                type = 'n'   # default: number
            pos = cell.attrib['r']
            value = cell.find(ns+'v')
            if value is not None:
                if type == 'n':
                    # Numbers as integers
                    value = float(value.text) if value.text is not None else None
                else:
                    value = value.text
            cols.append((pos, value))
            form = cell.find(ns+'f')
            if form is not None:
                text = str(form.text)
                forms[pos] = text.replace('\n', r'\n')
        rows.append((rcur, cols))
    # Conditional formatting
    # Styles first
    styles = os.path.join(tmpdir, 'xl', 'styles.xml')
    dxfs = []
    for d in ET.parse(styles).getroot().find(ns+'dxfs').iter(ns+'dxf'):
        fg = '-'
        font = d.find(ns+'font')
        if font:
            fg = font.find(ns+'color').attrib['rgb']
        bg = '-'
        fill = d.find(ns+'fill')
        if fill:
            fill = fill.find(ns+'patternFill')
            if fill:
                bg = fill.find(ns+'bgColor').attrib['rgb']
        dxfs.append(fg+'/'+bg)
    # Now the conditions
    cond_f = {}
    for c in root.iter(ns+'conditionalFormatting'):
        pos = c.attrib['sqref']
        conds = []
        for rule in c.iter(ns+'cfRule'):
            form = rule.find(ns+'formula')
            type = rule.attrib['type']
            if type == 'cellIs':
                txt = rule.attrib['operator'] + ' ' + form.text
            else:
                txt = '=' + form.text
            conds.append((txt, dxfs[int(rule.attrib['dxfId'])], int(rule.attrib['priority'])))
        cond_f[pos] = conds
    # Links are "Relationship"s
    links = {}
    urls = {}
    nr = '{http://schemas.openxmlformats.org/officeDocument/2006/relationships}'
    hlinks = root.find(ns+'hyperlinks')
    if hlinks:
        for r in hlinks.iter(ns+'hyperlink'):
            links[r.attrib['ref']] = r.attrib[nr+'id']
    # Read the strings
    strings = os.path.join(tmpdir, 'xl', 'sharedStrings.xml')
    # The cast to str is to support obsolete Python versions
    strs = [to_str(t.text) for t in ET.parse(strings).getroot().iter(ns+'t')]
    # Translate the links
    if links:
        # Read the relationships
        worksheet = os.path.join(tmpdir, 'xl', 'worksheets', '_rels', 'sheet'+str(sheet)+'.xml.rels')
        root = ET.parse(worksheet).getroot()
        rels = {}
        for r in root:
            rels[r.attrib['Id']] = r.attrib['Target']
        for pos, id in links.items():
            urls[pos] = rels[id]
    # Get the global definitions
    workbook = os.path.join(tmpdir, 'xl', 'workbook.xml')
    dnames = ET.parse(workbook).getroot().find(ns+'definedNames')
    vars = {}
    for dname in dnames.iter(ns+'definedName'):
        name = dname.attrib['name']
        vars[name] = dname.text
    name = os.path.basename(filename)
    pos_re = re.compile(r'(\D+)(\d+)')
    with open(os.path.join(TESTDIR, subdir, name + '.txt'), 'wt') as f:
        f.write('Variables:\n')
        for name, val in sorted(vars.items()):
            f.write(name + ' = ' + val + '\n')
        f.write('-'*80+'\n')
        used_cells = set()
        for r in rows:
            f.write('Row: ' + str(r[0]) + '\n')
            skip_next_col = False
            skip_str = None
            for col in r[1]:
                pos = col[0]
                used_cells.add(pos)
                m = pos_re.match(pos)
                cell = col[1]
                form = forms.get(pos)
                styles = cond_f.get(pos)
                if cell is None and form is None and styles is None:
                    continue
                f.write(' Col: ' + m.group(1) + '\n')
                if cell is not None:
                    f.write('   ')
                if isinstance(cell, str):
                    try:
                        text = '*NONE*' if cell == 'None' else strs[int(cell)]
                    except ValueError:
                        # Special cases where the text is there
                        text = cell
                    if text is None:
                        text = '*NONE*'
                    # Filter variable fields
                    if skip_next_col and (skip_str is None or skip_str in text):
                        f.write('*FILTERED*\n')
                        skip_next_col = False
                        continue
                    for filter in XLSX_FILTERS:
                        if text.startswith(filter[0]):
                            if filter[1] == 0:
                                text = '*FILTERED*'
                                break
                            skip_next_col = True
                            skip_str = filter[1]
                    url = urls.get(pos)
                    if url:
                        f.write('<a href="{}">{}</a>'.format(url, text))
                    else:
                        f.write('"' + text + '"')
                elif cell is not None:
                    # Python 2.7 str(float) has only 12 digits
                    # Forcing 16 fails in some cases
                    f.write("{:.12g}".format(cell))
                if cell is not None:
                    f.write('\n')
                if form:
                    f.write('  Formula: ' + form + '\n')
                if styles:
                    f.write('  Styles:\n')
                    for style in styles:
                        # f.write('  - {} -> {} ({})\n'.format(style[0], style[1], style[2]))
                        # The priority doesn't look really important and generates a lot of silly diffs
                        f.write('  - {} -> {}\n'.format(style[0], style[1]))
        # Orphan conditional formatting
        for pos in sorted(cond_f.keys()):
            if pos not in used_cells:
                f.write('Cell: ' + pos + '\n')
                for style in cond_f[pos]:
                    f.write(' - {} -> {}\n'.format(style[0], style[1]))
    shutil.rmtree(tmpdir)
    return True


def xlsx_to_csv(filename, subdir='result_test', price=True):
    res_csv = os.path.join(TESTDIR, subdir, filename + '.csv')
    out_xlsx = os.path.join(TESTDIR, filename + '.xlsx')
    # Convert to CSV
    logging.debug('Converting to CSV')
    cmd = ['xlsx2csv']
    if not price:
        cmd.append('--skipemptycolumns')
    cmd.append(out_xlsx)
    p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    # Filter it
    filter = r'\$ date|Prj date:.*\(file|kicost'
    if not price:
        filter += '|Total purchase'
    with open(res_csv, 'w') as f:
        p2 = subprocess.Popen(['egrep', '-i', '-v', '(' + filter + ')'], stdin=p1.stdout, stdout=f)
        p2.communicate()[0]


def check_diff(filename):
    ref = os.path.join(TESTDIR, 'expected_test', filename)
    res = os.path.join(TESTDIR, 'result_test', filename)
    cmd = ['diff', '-u', ref, res]
    log_running('Comparing', cmd)
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        logging.error(e.output.decode('utf-8'))
        raise


def run_test(name, inputs, output, extra=None, price=True, ret_err=0, config_file=None):
    if not os.path.isdir(TESTDIR + '/result_test'):
        os.mkdir(TESTDIR + '/result_test')
    if not os.path.isdir(TESTDIR + '/log_test'):
        os.mkdir(TESTDIR + '/log_test')
    # Always fake the currency rates
    os.environ['KICOST_CURRENCY_RATES'] = TESTDIR + '/currency_rates.xml'
    # Now choose between recording the KitSpace queries or fake them
    if price and ADD_QUERY_TO_KNOWN:
        os.environ['KICOST_LOG_HTTP'] = TESTDIR + '/kitspace_queries.txt'
        with open(TESTDIR + '/kitspace_queries.txt', 'at') as f:
            if len(inputs) == 1:
                f.write('# ' + inputs[0] + '\n')
            else:
                f.write('# ' + str(inputs) + '\n')
        server = None
    else:
        os.environ['KICOST_KITSPACE_URL'] = 'http://localhost:8000'
        os.environ['KICOST_OCTOPART_URL'] = 'http://localhost:8000'
        fo = open(TESTDIR + '/log_test/0server_stdout.log', 'at')
        fe = open(TESTDIR + '/log_test/0server_stderr.log', 'at')
        server = subprocess.Popen(TESTDIR + '/dummy-web-server.py', stdout=fo, stderr=fe)
    # Run KiCost
    cmd = ['src/kicost', '--debug', '10']
    if not price:
        cmd.append('--no_price')
    if extra:
        cmd.extend(extra)
    if config_file is not None:
        cmd.extend(['-c', TESTDIR + '/configs/' + config_file])
    else:
        cmd.extend(['-c', TESTDIR + '/configs/default.yaml'])
    out_xlsx = TESTDIR + '/' + output + '.xlsx'
    cmd.extend(['-o', out_xlsx])
    cmd.extend(['-wi'] + [TESTDIR + '/' + n for n in inputs])
    log_running('Testing', cmd)
    log_err = open(TESTDIR + '/log_test/' + output + '_error.log', 'w+t')
    log_out = open(TESTDIR + '/log_test/' + output + '_out.log', 'w+t')
    ret = subprocess.call(cmd, stderr=log_err, stdout=log_out)
    # Kill the server
    if server is not None:
        server.terminate()
        fo.close()
        fe.close()
    global last_err
    log_err.seek(0)
    last_err = log_err.read()
    log_err.close()
    log_out.close()
    # Check return value
    if ret_err != ret:
        logging.error('Failed test: ' + name)
        assert False, last_err
    # Convert to CSV/TXT
    if not ret_err:
        if CREATE_REF:
            xlsx_to_csv(output, 'expected_test', price)
            xlsx_to_txt(output, 'expected_test')
        else:
            xlsx_to_csv(output, 'result_test', price)
            check_diff(output + '.csv')
            xlsx_to_txt(output, 'result_test')
            check_diff(output + '.xlsx.txt')
    logging.info(output+' OK')


def run_test_check(name, inputs=None, output=None, extra=None, price=True, ret_err=0, config_file=None):
    logging.debug('Test name: ' + name)
    if inputs is None:
        inputs = name
    if isinstance(inputs, str):
        inputs = [inputs]
    if output is None:
        output = inputs[0]
        if output.endswith('.csv'):
            output = output[:-4]
    run_test(name, inputs, output, extra, price, ret_err, config_file)


def check_errors(errors):
    res = []
    global last_err
    for error in errors:
        m = re.search(error, last_err, re.MULTILINE)
        assert m is not None, error
        logging.debug('error match: `{}` (`{}`) OK'.format(error, m.group(0)))
        res.append(m)
    return res


def test_300_010():
    run_test_check('300-010')


def test_acquire_PWM_1():
    run_test_check('acquire-PWM', config_file='kitspace_no_cache.yaml')


def test_acquire_PWM_dk():
    name = 'acquire_PWM_dk'
    run_test_check(name, 'acquire-PWM', name, extra=['--include', 'digikey'], config_file='digikey.yaml')


def test_acquire_PWM_eur_dk():
    name = 'acquire_PWM_eur_dk'
    run_test_check(name, 'acquire-PWM', name, extra=['--include', 'digikey', '--currency', 'EUR'], config_file='digikey_eur.yaml')


def test_acquire_PWM_2():
    run_test_check('acquire-PWM_2', config_file='kitspace_no_cache.yaml')


def test_Aeronav_R():
    run_test_check('Aeronav_R')


def test_b3u():
    run_test_check('b3u_test')


def test_bbsram():
    run_test_check('bbsram')


def test_BoulderCreekMotherBoard():
    # This test doesn't have any kind of manf# or DISTRIBUTOR#
    run_test_check('BoulderCreekMotherBoard', price=False)


def test_CAN_Balancer():
    run_test_check('CAN Balancer')


def test_Decoder():
    run_test_check('Decoder')


def test_fitting():
    run_test_check('fitting_test')


def test_Indium_X2():
    run_test_check('Indium_X2')


def test_kc():
    run_test_check('kc-test')


def test_LedTest():
    run_test_check('LedTest')


def test_local_Indium_X2():
    run_test_check('local_Indium_X2')


def test_NF6X_TestBoard():
    run_test_check('NF6X_TestBoard')


def test_Receiver_1W():
    run_test_check('Receiver_1W')


def test_RPi():
    run_test_check('RPi-Test')


def test_RX_LR_lite():
    run_test_check('RX LR lite')


def test_safelink_receiver():
    run_test_check('safelink_receiver')


def test_single_component():
    run_test_check('single_component')


def test_StickIt_Hat_old():
    run_test_check('StickIt-Hat-old')


def test_StickIt_Hat_new():
    run_test_check('StickIt-Hat')


def test_StickIt_QuadDAC():
    run_test_check('StickIt-QuadDAC')


def test_StickIt_RotaryEncoder():
    # Tests an embedded price from Aliexpress
    run_test_check('StickIt-RotaryEncoder')


def test_subparts():
    run_test_check('subparts')


def test_subparts_err1():
    # Here we ask to repeat the manufacturer, but in the first position, nothing to repeat
    run_test_check('subparts_err1')


def test_1():
    run_test_check('test')


def test_2a():
    run_test_check('test2')


def test_2e():
    name = 'test_2e'
    run_test_check(name, 'test2', name, extra=['--currency', 'EUR'])


def test_3_():
    run_test_check('test3')


def test_Parts():
    run_test_check('TestParts')


def test_part_list_big():
    run_test_check('part_list_big.csv', config_file='kitspace_no_cache.yaml')


def test_part_list_small_hdr():
    run_test_check('part_list_small.csv', config_file='kitspace_no_cache.yaml')


def test_part_list_small_nohdr():
    run_test_check('part_list_small_nohdr.csv', config_file='kitspace_no_cache.yaml')


def test_multiproject_1():
    run_test_check('multiproject_1 (1 single)', 'multipart')
    run_test_check('multiproject_1 (2 single)', 'multipart2')
    run_test_check('multiproject_1', ['multipart', 'multipart2.xml'], 'multipart1+2')


def test_board_qty_1():
    # Check we can select 50 boards
    run_test_check('test_board_qty_1', 'test', 'board_qty_1', ['--board_qty', '50'])


def test_board_qty_2():
    # Check we can select 50 and 70 boards
    run_test_check('test_board_qty_2', ['multipart', 'multipart2.xml'], 'board_qty_2', ['--board_qty', '50', '70'])


def test_board_qty_3():
    # Check we can select 30 for all
    run_test_check('test_board_qty_3', ['multipart', 'multipart2.xml'], 'board_qty_3', ['--board_qty', '30'])


def test_variants_1():
    # This test doesn't have any kind of manf# or DISTRIBUTOR#
    test_name = 'variants_1'
    run_test_check(test_name, 'variants_1', price=False)
    run_test_check(test_name + '(test)', 'variants_1', 'variants_1_test', ['--variant', 'test'], price=False)
    run_test_check(test_name + '(production)', 'variants_1', 'variants_1_production', ['--variant', 'production'], price=False)
    run_test_check(test_name + '(default)', 'variants_1', 'variants_1_default', ['--variant', 'default'], price=False)


def test_variants_2():
    # This test is related to issue #474
    # Tests the same as test_manf_no_manf_num() but in a variant case
    # Note that we don't even have manf#, no price here
    test_name = 'variants_2'
    run_test_check(test_name, 'variants_2', price=False)
    run_test_check(test_name + '(variant1)', 'variants_2', 'variants_2_variant1', ['--variant', 'variant1'], price=False)


def test_variants_3():
    # This test doesn't have any kind of manf# or DISTRIBUTOR#
    # Tests some variant overwrites
    test_name = 'variants_3'
    run_test_check(test_name, 'variants_3', price=False)
    # Run a test with parameter "variant1"
    run_test_check(test_name + '(variant1)', 'variants_3', 'variants_3_variant1', ['--variant', '^(variant1)$', '--fields', 'Comment'], price=False)


def test_user_fields_1():
    run_test_check('user_fields_1', '300-010', 'user_fields_1', extra=['--fields', "Resistance", "Capacitance", "Voltage", "Tolerance"])


def test_complex_multipart():
    # This testcase has to be updated once multipart custom pricing has been better defined
    test_name = 'complex_multipart'
    fields = ['S1MN', 'S1PN', 'S2MN', 'S2PN']
    run_test_check(test_name, 'complex_multipart', extra=['--split_extra_fields'] + fields + ['-f'] + fields, price=True)


def test_include_1():
    # Explicitly request digikey and mouser
    run_test_check('include_1', 'fitting_test', 'include_1', extra=['--include', 'digikey', 'mouser'])


def test_exclude_1():
    # Implicitly request digikey and mouser
    run_test_check('exclude_1', 'fitting_test', 'exclude_1', extra=['--exclude', 'arrow', 'farnell', 'lcsc', 'newark', 'rs', 'tme'])


def test_scrape_over_1():
    # Data from the fields is added to the web-scraped data
    run_test_check('scrape_over')


def test_scrape_over_2():
    # Data from the fields relaces the web-scraped data.
    # For this we exclude the distributor with --exclude
    run_test_check('scrape_over_2', 'scrape_over', 'scrape_over_2', extra=['--exclude', 'rs'], config_file='kitspace_no_cache.yaml')


def test_manf_no_manf_num():
    # Two similar parts, but from different manufacturer and no manf#
    # Issue #474
    run_test_check('manf_no_manf_num')


def test_parts_and_comments():
    # Similar to test_no_empty_overwrite, tests all possible manf# aliases
    run_test_check('parts_and_comments', extra=['--group_fields', 'h', 'comment',
                   '--no_collapse', '-f', 'comment', 'S1MN', 'S1PN', 'S2MN', 'S2PN'], price=False)


def test_group_1():
    # Similar to test_no_empty_overwrite, tests all possible manf# aliases
    run_test_check('group_1_group_fields', 'group_1', output='group_1_group_fields',
                   extra=['--group_fields', 'h', 'comment', '--no_collapse', '-f', 'comment', 'S1MN', 'S1PN', 'S2MN', 'S2PN'],
                   price=False)
    run_test_check('group_1_ignore_comment', 'group_1', output='group_1_ignore_comment',
                   extra=['--ignore_fields', 'h', 'comment', '--no_collapse', '-f', 'comment', 'S1MN', 'S1PN', 'S2MN', 'S2PN'],
                   price=False)


def test_423():
    # Test for issue #423
    # This test checks that we interpret a numeric manf# code as a string
    # The "OK" tests uses the real manf#
    # The "Wrong" tests uses an invalid value, reported in #423
    # Checking how it looks in the spreadsheet software needs manual inspect, but currently we use "write_string".
    # Any "scientic notation" is a bug in the spreadsheet software. MS Excel does it right.
    run_test_check('Test 423 CSV Ok', 'test_423_ok.csv', 'test_423_csv_ok')
    run_test_check('Test 423 CSV Wrong', 'test_423_wrong.csv', 'test_423_csv_wrong')
    run_test_check('Test 423 XML Ok', 'test_423_ok', 'test_423_xml_ok')
    run_test_check('Test 423 XML Wrong', 'test_423_wrong', 'test_423_xml_wrong')


def disabled_test_sub_part_group_propagate_266():
    # Test Issue #266
    #  SubPart manf# field should also propagate
    run_test_check('SubPartGroupTest_266', price=False)


def test_no_empty_overwrite():
    # Test some cases where we overwrite a field using an alias (i.e. mnp changes manf#)
    # See discusion on #471
    run_test_check('no_empty_overwrite', price=False)


def test_wrong_pricing():
    # File with errors in the pricing field
    run_test_check('wrong_pricing', extra=['--include', 'arrow', '--exclude', 'arrow'])
    check_errors([r'Malformed pricing number(.*)STK1', r'Malformed pricing entry(.*)PCB1'])


def test_wrong_currency():
    # File with a wrong currency
    run_test_check('wrong_currency', extra=['--include', 'arrow', '--exclude', 'arrow'], ret_err=ERR_FIELDS)
    check_errors([r'XXX is not a supported currency in STK1'])


def test_rare_refs_collapse():
    # File with a wrong currency
    name = 'rare_refs_collapse'
    run_test_check(name, 'rare_refs', name, price=False)


def test_rare_refs_no_collapse():
    # File with a wrong currency, disable collapse
    name = 'rare_refs_no_collapse'
    run_test_check(name, 'rare_refs', name, extra=['--no_collapse'], price=False)


def test_octopart_1p():
    name = 'octopart_1'
    run_test_check(name + 'p', name, name + 'p', extra=['--octopart_key', OCTOPART_KEY, '--octopart_level', '4p'], config_file='octopart_no_cache.yaml')


def test_octopart_1n():
    name = 'octopart_1'
    run_test_check(name + 'n', name, name + 'n', config_file='octopart_no_cache.yaml')


def test_octopart_1_ambi():
    name = 'octopart_1_ambi'
    run_test_check(name, extra=['--octopart_key', OCTOPART_KEY, '--octopart_level', '4p'], config_file='octopart_no_cache.yaml')
    check_errors([r'Using "Adafruit Industries" for manf#="4062"', r'Ambiguous manf#="4062" please use manf to select the right one, choices:'])


def test_octopart_2n():
    name = 'octopart_2'
    run_test_check(name + 'n', name, name + 'n', extra=['--disable_api', 'KitSpace'], config_file='octopart.yaml')


def test_337():
    # Test for issue #337
    run_test_check('test_337_UserFieldCombining', extra=['--field', 'Supplier'], price=False)


def test_mouser_1():
    test_name = 'mouser_1'
    fields = ['S1MN', 'S1PN', 'S2MN', 'S2PN']
    extra = ['--split_extra_fields'] + fields + ['-f'] + fields + ['--include', 'mouser']
    run_test_check(test_name, 'complex_multipart', test_name, extra=extra, config_file='mouser.yaml')


def test_element14_1():
    test_name = 'element14_1'
    extra = ['--include', 'farnell', 'newark']
    run_test_check(test_name, 'safelink_receiver', test_name, extra=extra, config_file='element14.yaml')


def test_tme_1():
    test_name = 'tme_1'
    extra = ['--include', 'tme']
    run_test_check(test_name, 'safelink_receiver', test_name, extra=extra, config_file='tme.yaml')


class TestKicost(unittest.TestCase):

    def setUp(self):
        pass

    def test_something(self):
        pass

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



# Copyright Jonathan Hartley 2013. BSD 3-Clause license, see LICENSE file.
'''
This module generates ANSI character codes to printing colors to terminals.
See: http://en.wikipedia.org/wiki/ANSI_escape_code
'''
import sys
import os

CSI = '\033['
OSC = '\033]'
BEL = '\a'
is_a_tty = sys.stderr.isatty() and os.name == 'posix'


def code_to_chars(code):
    return CSI + str(code) + 'm' if is_a_tty else ''


def set_title(title):
    return OSC + '2;' + title + BEL


# def clear_screen(mode=2):
#     return CSI + str(mode) + 'J'


# def clear_line(mode=2):
#     return CSI + str(mode) + 'K'


class AnsiCodes(object):
    def __init__(self):
        # the subclasses declare class attributes which are numbers.
        # Upon instantiation we define instance attributes, which are the same
        # as the class attributes but wrapped with the ANSI escape sequence
        for name in dir(self):
            if not name.startswith('_'):
                value = getattr(self, name)
                setattr(self, name, code_to_chars(value))


# class AnsiCursor(object):
#     def UP(self, n=1):
#         return CSI + str(n) + 'A'
#
#     def DOWN(self, n=1):
#         return CSI + str(n) + 'B'
#
#     def FORWARD(self, n=1):
#         return CSI + str(n) + 'C'
#
#     def BACK(self, n=1):
#         return CSI + str(n) + 'D'
#
#     def POS(self, x=1, y=1):
#         return CSI + str(y) + ';' + str(x) + 'H'


class AnsiFore(AnsiCodes):
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 39

    # These are fairly well supported, but not part of the standard.
    LIGHTBLACK_EX = 90
    LIGHTRED_EX = 91
    LIGHTGREEN_EX = 92
    LIGHTYELLOW_EX = 93
    LIGHTBLUE_EX = 94
    LIGHTMAGENTA_EX = 95
    LIGHTCYAN_EX = 96
    LIGHTWHITE_EX = 97


class AnsiBack(AnsiCodes):
    BLACK = 40
    RED = 41
    GREEN = 42
    YELLOW = 43
    BLUE = 44
    MAGENTA = 45
    CYAN = 46
    WHITE = 47
    RESET = 49

    # These are fairly well supported, but not part of the standard.
    LIGHTBLACK_EX = 100
    LIGHTRED_EX = 101
    LIGHTGREEN_EX = 102
    LIGHTYELLOW_EX = 103
    LIGHTBLUE_EX = 104
    LIGHTMAGENTA_EX = 105
    LIGHTCYAN_EX = 106
    LIGHTWHITE_EX = 107


class AnsiStyle(AnsiCodes):
    BRIGHT = 1
    DIM = 2
    NORMAL = 22
    RESET_ALL = 0


Fore = AnsiFore()
Back = AnsiBack()
Style = AnsiStyle()
# Cursor = AnsiCursor()



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (c) 2020-2022 Salvador E. Tropea
# Copyright (c) 2020-2022 Instituto Nacional de Tecnología Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft/'
__company__ = 'INTI-CMNB - Argentina'

import copy
import os
import logging
import tqdm
# QueryCache dependencies:
import pickle
import time
from .. import DEBUG_HTTP_HEADERS, DEBUG_HTTP_RESPONSES, DEBUG_OBSESSIVE
from ..global_vars import DEFAULT_CURRENCY, BASE_OP_TYPES, W_NOAPI
from .distributors_info import distributors_info

__all__ = ['distributor_class']


class TqdmLoggingHandler(logging.Handler):
    '''Overload the class to write the logging through the `tqdm`.'''
    def __init__(self, stream, level=logging.NOTSET):
        super(self.__class__, self).__init__(level)
        self.stream = stream

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg, file=self.stream)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


class QueryCache(object):
    ''' Components queries cache implementation '''
    def __init__(self, path, ttl):
        self.path = path
        self.ttl_min = ttl*24*60
        self.suffix = ''

    def get_name(self, prefix, name):
        return os.path.join(self.path, prefix + '_' + name.replace('/', '_') + self.suffix + ".dat")

    def save_results(self, prefix, name, results):
        ''' Saves the results to the cache '''
        with open(self.get_name(prefix, name), "wb") as fh:
            pickle.dump(results, fh, protocol=2)

    def load_results(self, prefix, name):
        ''' Loads the results from the cache, must be implemented by KiCost '''
        file = self.get_name(prefix, name)
        distributor_class.logger.log(DEBUG_OBSESSIVE, 'Looking in cache '+file)
        if not os.path.isfile(file):
            return None, False
        mtime = os.path.getmtime(file)
        ctime = time.time()
        dif_minutes = int((ctime-mtime)/60)
        if self.ttl_min < 0 or (self.ttl_min > 0 and dif_minutes <= self.ttl_min):
            with open(file, "rb") as fh:
                result = pickle.loads(fh.read())
            distributor_class.logger.log(DEBUG_OBSESSIVE, 'Found in cache '+str(result))
            return result, True
        # Cache expired
        return None, False


class distributor_class(object):
    registered = []
    priorities = []
    logger = None
    # distributor_dict contains the available distributors.
    # The distributors are added by the api_*/dist_*/scrape_* modules.
    # The information of each distributor is copied from distributors_info
    # Some modules can add new distributors, not found on distributors_info (from data in the fields)
    # The list of *used* distributors is handled separately.
    distributor_dict = {}
    label2name = {}
    # Options supported by this API
    config_options = {}

    @staticmethod
    def register(api, priority):
        index = 0
        for idx, prio in enumerate(distributor_class.priorities):
            index = idx
            if prio < priority:
                break
        else:
            index += 1
        distributor_class.registered.insert(index, api)
        distributor_class.priorities.insert(index, priority)

    @staticmethod
    def update_distributors(parts, distributors):
        """ This is used by the Local distributors mechanism to discover user defined distributors """
        pass

    @staticmethod
    def get_dist_parts_info(parts, distributors, currency=DEFAULT_CURRENCY):
        ''' Get the parts info using the modules API/Scrape/Local.'''
        from .log__ import debug_overview
        debug_overview('Starting to search using distributors: {}'.format(distributors))
        # Discover user defined distributors
        for api in distributor_class.registered:
            if api.enabled:
                api.update_distributors(parts, distributors)
        debug_overview('Distributors after local discovery: {}'.format(distributors))
        # Now look for the parts
        remaining = set(distributors)
        for api in distributor_class.registered:
            debug_overview('Considering: {} {}'.format(api.name, api.api_distributors))
            if api.enabled and len(remaining.intersection(api.api_distributors)):
                if api.url:
                    url = ' ({})'.format(api.url)
                else:
                    url = ''
                distributor_class.logger.info('- {} [{}]{}'.format(api.name, api.type, url))
                solved = api.query_part_info(parts, list(remaining), currency)
                if solved:
                    remaining -= solved
                debug_overview('Distributors solved {}, remaining {}'.format(solved, remaining))

    @classmethod
    def init_dist_dict(cls):
        if cls.enabled and cls.type != 'local':   # Local distributors are collected from the components
            distributor_class.add_distributors(cls.api_distributors)

    @staticmethod
    def main_init_dist_dict():
        ''' Initialize and update the dictionary of the registered distributors classes.'''
        # Clear distributor_dict, then let all distributor modules recreate their entries.
        distributor_class.distributor_dict = {}
        distributor_class.label2name = {}
        for api in distributor_class.registered:
            api.init_dist_dict()

    @staticmethod
    def get_distributors_iter():
        return distributor_class.distributor_dict.keys()

    @staticmethod
    def add_distributors(dists):
        ''' Adds a copy of the distributor info to the supported '''
        for dist in dists:
            # Here we copy the available distributors from distributors_info
            # We use a copy so they can be restored just calling this init again
            data = distributors_info[dist]
            distributor_class.distributor_dict[dist] = copy.deepcopy(data)
            distributor_class.label2name[data.label.name.lower()] = dist

    @staticmethod
    def add_distributor(name, data):
        ''' Adds a distributor to the list of supported '''
        distributor_class.distributor_dict[name] = data
        distributor_class.label2name[data.label.name.lower()] = name

    @staticmethod
    def get_distributor_template(name):
        ''' Get a copy of the distributor info from the original structure.
            Used internaly from the API to add distributors derived from others. '''
        return copy.deepcopy(distributors_info[name])

    @staticmethod
    def get_distributor_info(name):
        ''' Gets all the information about a supported distributor.
            This information comes from the list collected from the APIs, not from the fixed template. '''
        return distributor_class.distributor_dict[name]

    @staticmethod
    def log_request(url, data):
        distributor_class.logger.log(DEBUG_HTTP_HEADERS, 'URL ' + url + ' query:')
        distributor_class.logger.log(DEBUG_HTTP_HEADERS, data)
        if os.environ.get('KICOST_LOG_HTTP'):
            with open(os.environ['KICOST_LOG_HTTP'], 'at') as f:
                f.write(data + '\n')

    @staticmethod
    def log_response(response):
        distributor_class.logger.log(DEBUG_HTTP_RESPONSES, response.text)
        distributor_class.logger.log(DEBUG_HTTP_RESPONSES, 'Status Code: <{}>'.format(response.status_code))
        if os.environ.get('KICOST_LOG_HTTP'):
            with open(os.environ['KICOST_LOG_HTTP'], 'at') as f:
                f.write(response.text + '\n')

    @staticmethod
    def _get_api(api):
        # We currently assume the API is registered
        return next((x for x in distributor_class.registered if x.name == api), None)

    @staticmethod
    def configure_apis(options):
        ''' Configure all APIs. options is a dict API -> api_options '''
        for api_name, ops in options.items():
            api = distributor_class._get_api(api_name)
            if api is not None:
                api.configure(ops)

    @staticmethod
    def set_api_status(api_name, enabled):
        ''' Enable/Disable a particular API '''
        api = distributor_class._get_api(api_name)
        if api:
            api.enabled = enabled
        else:
            distributor_class.logger.warning(W_NOAPI+'No API registered as `{}`'.format(api_name))

    @staticmethod
    def get_api_status(api):
        ''' Find if an API is enabled '''
        return distributor_class._get_api(api).enabled

    @staticmethod
    def from_environment(options, overwrite):
        ''' Default configuration from the environment. Just nothing. '''
        pass

    @staticmethod
    def _set_from_env(key, value, options, overwrite, d_types=None):
        ''' Helper function to implement `from_environment`. '''
        if value is not None and (overwrite or key not in options):
            if d_types:
                # If we know the valid data type for the value ensure it
                tp = d_types.get(key, None)
                tp = BASE_OP_TYPES.get(key, tp)
                if not isinstance(tp, type):
                    # Solve the case where more than one data type is allowed
                    tp = tp[0]
                # This is a cast
                value = tp(value)
            options[key] = value

    @staticmethod
    def configure_from_environment(options, overwrite):
        ''' Configure all APIs using environment variables.
            If overwrite is True the API should replace the current option '''
        for api in distributor_class.registered:
            api.from_environment(options[api.name], overwrite)

    # Abstract methods, implemented in distributor specific modules.
    @staticmethod
    def query():
        '''Send query to server and return results.'''
        raise NotImplementedError()

    @staticmethod
    def query_part_info():
        ''' Get the parts info of one distributor class.'''
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Max Maisel / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Libraries.
import json
import requests
import re
import os
import sys
import difflib
from collections import Counter
if sys.version_info[0] < 3:
    from urllib import quote_plus
else:
    from urllib.parse import quote_plus

# KiCost definitions.
from .. import KiCostError, DistData, ERR_SCRAPE, W_ASSQTY, W_AMBIPN, W_APIFAIL
# Distributors definitions.
from .distributor import distributor_class, QueryCache
from .log__ import debug_overview, debug_obsessive, warning

# Author information.
__author__ = 'XESS Corporation'
__webpage__ = 'info@xess.com'

OCTOPART_MAX_PARTBYQUERY = 20  # Maximum part list length to one single query.

__all__ = ['api_octopart']


class QueryStruct(object):
    def __init__(self, id, kind, code, part):
        self.id = str(id)  # ID number for this query
        self.kind = kind   # mpn/sku
        self.code = code   # manf_code/sku
        self.part = part   # Pointer to the part.
        self.result = None
        self.loaded = False


class api_octopart(distributor_class):
    name = 'Octopart'
    type = 'api'
    enabled = False
    url = 'https://octopart.com/'  # Web site API information.
    # Options supported by this API
    config_options = {'key': str, 'level': int, 'extended': bool}
    api_level = 4
    # Include specs and datasheets. Only in the Pro plan.
    extended = False
    cache = None

    API_KEY = None
    api_distributors = ['arrow', 'digikey', 'farnell', 'lcsc', 'mouser', 'newark', 'rs', 'tme']
    DIST_TRANSLATION = {  # Distributor translation. Just a few supported.
                        'Arrow Electronics': 'arrow',
                        'Digi-Key': 'digikey',
                        'Farnell': 'farnell',
                        'Mouser': 'mouser',
                        'Newark': 'newark',
                        'RS Components': 'rs',
                        'TME': 'tme',
                        'LCSC': 'lcsc',
                       }
    # Dict to translate KiCost field names into Octopart distributor names
    KICOST2OCTOPART_DIST = {v: k for k, v in DIST_TRANSLATION.items()}

    @staticmethod
    def configure(ops):
        cache_ttl = 7
        cache_path = None
        for k, v in ops.items():
            if k == 'key':
                api_octopart.API_KEY = v
                if 'enable' not in ops:
                    # If not explicitly disabled then enable it
                    api_octopart.enabled = True
            elif k == 'enable':
                api_octopart.enabled = v
            elif k == 'extended':
                api_octopart.extended = v
            elif k == 'level':
                api_octopart.api_level = v
            elif k == 'cache_ttl':
                cache_ttl = v
            elif k == 'cache_path':
                cache_path = v
        api_octopart.cache = QueryCache(cache_path, cache_ttl)
        if api_octopart.enabled and api_octopart.API_KEY is None:
            warning(W_APIFAIL, "Can't enable Octopart without a `key`")
            api_octopart.enabled = False
        debug_obsessive('Octopart API configured to enabled {} key {} level {} extended {}'.
                        format(api_octopart.enabled, api_octopart.API_KEY, api_octopart.api_level, api_octopart.extended))

    @staticmethod
    def query(query):
        """Send query to Octopart and return results."""
        # url = 'http://octopart.com/api/v3/parts/match'
        # payload = {'queries': json.dumps(query), 'include\[\]': 'specs', 'apikey': token}
        # response = requests.get(url, params=payload)
        data = 'queries=['+', '.join(query)+']'
        if api_octopart.API_KEY:
            if api_octopart.api_level == 3:
                url = 'http://octopart.com/api/v3/parts/match'
            else:
                url = 'http://octopart.com/api/v4/rest/parts/match'
            data += '&apikey=' + api_octopart.API_KEY
        else:  # Not working 2021/04/28:
            url = 'https://temp-octopart-proxy.kitspace.org/parts/match'
        # Allow changing the URL for debug purposes
        try:
            url = os.environ['KICOST_OCTOPART_URL']
        except KeyError:
            pass
        if api_octopart.extended:
            data += '&include[]=specs'
            data += '&include[]=datasheets'
        distributor_class.log_request(url, data)
        response = requests.get(url + '?' + data)
        distributor_class.log_response(response)
        if response.status_code == 200:  # Ok
            results = json.loads(response.text).get('results')
            return results
        elif response.status_code == 400:  # Bad request
            raise KiCostError('Octopart missing apikey.', ERR_SCRAPE)
        elif response.status_code == 404:  # Not found
            raise KiCostError('Octopart server not found.', ERR_SCRAPE)
        elif response.status_code == 403 or 'Invalid API key' in response.text:
            raise KiCostError('Octopart KEY invalid, register one at "https://www.octopart.com".', ERR_SCRAPE)
        elif response.status_code == 429:  # Too many requests
            raise KiCostError('Octopart request limit reached.', ERR_SCRAPE)
        else:
            raise KiCostError('Octopart error: ' + str(response.status_code), ERR_SCRAPE)

    @staticmethod
    def sku_to_mpn(sku):
        """Find manufacturer part number associated with a distributor SKU."""
        part_query = ['{"reference": "1", "sku": "'+quote_plus(sku)+'"}']
        results = api_octopart.query(part_query)
        if not results:
            return None
        result = results[0]
        mpns = [item['mpn'] for item in result['items']]
        if not mpns:
            return None
        if len(mpns) == 1:
            return mpns[0]
        mpn_cnts = Counter(mpns)
        return mpn_cnts.most_common(1)[0][0]  # Return the most common MPN.

    @staticmethod
    def skus_to_mpns(parts, distributors):
        """Find manufaturer's part number for all parts with just distributor SKUs."""
        for i, part in enumerate(parts):

            # Skip parts that already have a manufacturer's part number.
            if part.fields.get('manf#'):
                continue

            # Get all the SKUs for this part.
            skus = list(
                set([part.fields.get(dist + '#', '') for dist in distributors]))
            skus = [sku for sku in skus
                    if sku not in ('', None)]  # Remove null SKUs.

            # Skip this part if there are no SKUs.
            if not skus:
                continue

            # Convert the SKUs to manf. part numbers.
            mpns = [api_octopart.sku_to_mpn(sku) for sku in skus]
            mpns = [mpn for mpn in mpns
                    if mpn not in ('', None)]  # Remove null manf#.

            # Skip assigning manf. part number to this part if there aren't any to assign.
            if not mpns:
                continue

            # Assign the most common manf. part number to this part.
            mpn_cnts = Counter(mpns)
            part.fields['manf#'] = mpn_cnts.most_common(1)[0][0]

    @staticmethod
    def get_part_info(queries):
        """Query Octopart for quantity/price info."""
        only_query = ['{"reference": "'+q.id+'", "'+q.kind+'": "'+quote_plus(q.code)+'"}' for q in queries]
        results = api_octopart.query(only_query)
        # Copy the results to the queries
        q_hash = {q.id: q for q in queries}
        for r in results:
            query = q_hash[r['reference']]
            query.result = r
            api_octopart.cache.save_results(query.kind, query.code, r)

    @staticmethod
    def fill_part_info(queries, distributors, solved, currency='USD'):
        ''' Place the results into the parts list. '''
        # Translate from Octopart distributor names to the names used internally by kicost.
        dist_xlate = api_octopart.DIST_TRANSLATION
        # List of desired distributors in native format
        native_dists = set((api_octopart.KICOST2OCTOPART_DIST[d] for d in distributors))
        # Currency priority: 1: User specified, 2: USD, 3: EUR
        currency_prio = [currency]
        if currency != 'USD':
            currency_prio.append('USD')
        if currency != 'EUR':
            currency_prio.append('EUR')
        # Loop through the response to the query and enter info into the parts list.
        for query in queries:
            result = query.result
            part = query.part
            # Each result can match one or more components from different manufacturers
            # Take only the items with useful offers
            useful_items = []
            for item in result['items']:
                for offer in item['offers']:
                    if offer['seller']['name'] in native_dists:
                        useful_items.append(item)
                        break
            # debug_obsessive(str(result))
            # debug_obsessive(str(useful_items))
            # If more than one select the right one
            if len(useful_items) > 1:
                # List of possible manufacturers
                # TODO: Can we get more than one hit for the same manf?
                manufacturers = {it['manufacturer']['name'].lower(): it for it in useful_items}
                # Is the manf included?
                manf = part.fields.get('manf', 'none').lower()
                item = manufacturers.get(manf)
                if not item:
                    if manf == 'none':
                        item = useful_items[0]
                    else:
                        best_match = difflib.get_close_matches(manf, manufacturers.keys())[0]
                        item = manufacturers[best_match]
                    mpn = item['mpn']
                    warning(W_AMBIPN, 'Using "{}" for manf#="{}"'.format(item['manufacturer']['name'], mpn))
                    warning(W_AMBIPN, 'Ambiguous manf#="{}" please use manf to select the right one, choices: {}'.format(
                            mpn, list(manufacturers.keys())))
            else:
                if len(useful_items):
                    item = useful_items[0]
                else:
                    # No hits, skip
                    continue
            if api_octopart.extended:
                # Assign the lifecycle status 'obsolete' (others possible: 'active' and 'not recommended for new designs') but not used.
                try:
                    # API v4 (production, eol, nrnd, ...) we take the first word
                    part.lifecycle = item['specs']['lifecyclestatus']['value'][0].lower().split(' ')[0]
                except KeyError:
                    try:
                        # API v3
                        part.lifecycle = item['specs']['lifecycle_status']['value'][0].lower()
                    except KeyError:
                        # No lifecyclestatus (current name) nor lifecycle_status (old name)
                        pass
                # Take the datasheet provided by the distributor. This will by used
                # in the output spreadsheet if not provide any in the BOM/schematic.
                # This will be signed in the file.
                try:
                    part.datasheet = item['datasheets'][0]['url']
                except (KeyError, IndexError):
                    # No datasheet key (KeyError) or empty (IndexError)
                    pass
                # Misc data collected, currently not used inside KiCost
                part.update_specs({code: (info['metadata']['name'], ', '.join(info['value'])) for code, info in item['specs'].items()})
            # Loop through the offers from various dists for this particular part.
            for offer in item['offers']:
                # Get the distributor who made the offer and add their
                # price/qty info to the parts list if its one of the accepted distributors.
                dist = dist_xlate.get(offer['seller']['name'], '')
                if dist not in distributors:
                    # Unknown or excluded seller
                    continue
                price_tiers = {}
                part_qty_increment = float("inf")
                # Get the DistData for this distributor
                dd = part.dd.get(dist, DistData())
                # Get pricing information from this distributor.
                prices = offer['prices']
                if prices:
                    for curr in currency_prio:
                        if curr in prices:
                            dd.currency = curr
                            price_l = prices[curr]
                            break
                    else:
                        # Use the first entry
                        dd.currency, price_l = next(iter(prices.items()))
                    price_tiers = {qty: float(price) for qty, price in price_l}
                    # Combine price lists for multiple offers from the same distributor
                    # to build a complete list of cut-tape and reeled components.
                    dd.price_tiers.update(price_tiers)
                    # Compute the quantity increment between the lowest two prices.
                    # This will be used to distinguish the cut-tape from the reeled components.
                    try:
                        part_break_qtys = sorted(price_tiers.keys())
                        part_qty_increment = part_break_qtys[1] - part_break_qtys[0]
                    except IndexError:
                        # This will happen if there are not enough entries in the price/qty list.
                        # As a stop-gap measure, just assign infinity to the part increment.
                        # A better alternative may be to examine the packaging field of the offer.
                        pass
                # Select the part SKU, web page, and available quantity.
                # Each distributor can have different stock codes for the same part in different
                # quantities / delivery package styles: cut-tape, reel, ...
                # Therefore we select and overwrite a previous selection if one of the
                # following conditions is met:
                #   1. We don't have a selection for this part from this distributor yet.
                #   2. The MOQ is smaller than for the current selection.
                #   3. The part_qty_increment for this offer smaller than that of the existing selection.
                #      (we prefer cut-tape style packaging over reels)
                #   4. For DigiKey, we can't use part_qty_increment to distinguish between
                #      reel and cut-tape, so we need to look at the actual DigiKey part number.
                #      This procedure is made by the definition `distributors_info[dist]['ignore_cat#_re']`
                #      at the distributor profile.
                if not dd.part_num:
                    qty_avail = dd.qty_avail
                    in_stock_quantity = offer.get('in_stock_quantity')
                    if not qty_avail or (in_stock_quantity and qty_avail < in_stock_quantity):
                        # Keep the information with more availability.
                        dd.qty_avail = in_stock_quantity
                    moq = dd.moq
                    moq_offer = offer.get('moq')
                    if not moq or (moq_offer and moq > moq_offer):
                        # Save the link, stock code, ... of the page for minimum purchase.
                        dd.moq = moq_offer  # Minimum order qty.
                        dd.part_num = offer.get('sku')
                        dd.url = offer.get('product_url')
                        dd.qty_increment = part_qty_increment
                # Otherwise, check qty increment and see if its the smallest for this part & dist.
                elif part_qty_increment < dd.qty_increment:
                    # This part looks more like a cut-tape version, so
                    # update the SKU, web page, and available quantity.
                    qty_avail = dd.qty_avail
                    in_stock_quantity = offer.get('in_stock_quantity')
                    if not qty_avail or (in_stock_quantity and qty_avail < in_stock_quantity):
                        # Keep the information with more availability.
                        dd.qty_avail = in_stock_quantity
                    # Check for a valid SKU
                    dist_part_num = offer.get('sku', '')
                    ign_stock_code = distributor_class.get_distributor_info(dist).ignore_cat
                    valid_part = not (ign_stock_code and re.match(ign_stock_code, dist_part_num))
                    moq_offer = offer.get('moq')
                    if (valid_part and (not dd.part_num or (part_qty_increment < dd.qty_increment) or
                                        (not dd.moq or (moq_offer and dd.moq > moq_offer)))):
                        # Save the link, stock code, ... of the page for minimum purchase.
                        dd.moq = moq_offer  # Minimum order qty.
                        dd.part_num = dist_part_num
                        dd.url = offer.get('product_url')
                        dd.qty_increment = part_qty_increment
                # Update the DistData for this distributor
                part.dd[dist] = dd
                # We have data for this distributor
                solved.add(dist)

    @staticmethod
    def query_part_info(parts, distributors, currency):
        """Fill-in the parts with price/qty/etc info from Octopart."""
        debug_overview('# Getting part data from Octopart...')

        # Get the valid distributors names used by them part catalog
        # that may be index by Octopart. This is used to remove the
        # local distributors and future not implemented in the Octopart
        # definition.
        # Note: The user can use --exclude and define it with fields.
        distributors_octopart = [d for d in distributors if distributor_class.get_distributor_info(d).is_web()
                                 and d in api_octopart.api_distributors]

        # Collect all the queries
        queries = []
        for i, part in enumerate(parts):
            # Create an Octopart query using the manufacturer's part number or distributor SKU.
            manf_code = part.fields.get('manf#')
            if manf_code:
                query = QueryStruct(i, "mpn", manf_code, part)
            else:
                # No MPN, so use the first distributor SKU that's found.
                for octopart_dist_sku in distributors_octopart:
                    sku = part.fields.get(octopart_dist_sku + '#', '')
                    if sku:
                        break
                if not sku:
                    # No MPN or SKU, so skip this part.
                    continue
                # Create the part query using SKU matching.
                query = QueryStruct(i, "sku", sku, part)

                # Because was used the distributor (enrolled at Octopart list)
                # despite the normal 'manf#' code, take the sub quantity as
                # general sub quantity of the current part.
                try:
                    part.fields['manf#_qty'] = part.fields[octopart_dist_sku + '#_qty']
                    warning(W_ASSQTY, "Associated {q} quantity to '{r}' due \"{f}#={q}:{c}\".".format(
                            q=part.fields[octopart_dist_sku + '#_qty'], r=part.refs,
                            f=octopart_dist_sku, c=part.fields[octopart_dist_sku+'#']))
                except KeyError:
                    pass
            # Add query for this part to the list of part queries.
            queries.append(query)

        n_queries = len(queries)
        debug_overview('Queries {}'.format(n_queries))
        if not n_queries:
            return

        # Try to solve the queries from the cache
        unsolved = []
        for query in queries:
            # Look in the cache
            query.result, query.loaded = api_octopart.cache.load_results(query.kind, query.code)
            if not query.loaded:
                unsolved.append(query)

        # Solve the rest from the site
        n_unsolved = len(unsolved)
        debug_overview('Cached entries {}'.format(n_queries-n_unsolved))
        if n_unsolved:
            # Setup progress bar to track progress of server queries.
            progress = distributor_class.progress(n_unsolved, distributor_class.logger)

            # Slice the queries into batches of the largest allowed size and gather
            # the part data for each batch.
            for i in range(0, n_unsolved, OCTOPART_MAX_PARTBYQUERY):
                slc = slice(i, i+OCTOPART_MAX_PARTBYQUERY)
                api_octopart.get_part_info(unsolved[slc])
                progress.update(len(unsolved[slc]))

            # Done with the scraping progress bar so delete it or else we get an
            # error when the program terminates.
            progress.close()

        # Transfer the results
        solved_dist = set()
        api_octopart.fill_part_info(queries, distributors_octopart, solved_dist, currency)
        return solved_dist

    @staticmethod
    def from_environment(options, overwrite):
        ''' Configuration from the environment. '''
        # Configure the module from the environment
        # The command line will overwrite it using set_options()
        key = os.getenv('KICOST_OCTOPART_KEY_V3')
        if key:
            api_octopart._set_from_env('key', key, options, overwrite)
            api_octopart._set_from_env('enable', True, options, overwrite)
            api_octopart._set_from_env('level', 3, options, overwrite)
        else:
            key = os.getenv('KICOST_OCTOPART_KEY_V4')
            if key:
                api_octopart._set_from_env('key', key, options, overwrite)
                api_octopart._set_from_env('enable', True, options, overwrite)
                api_octopart._set_from_env('level', 4, options, overwrite)
            elif os.environ.get('KICOST_OCTOPART'):
                # Currently this isn't useful, you can't do anything without a key.
                # This is just in case we get a proxy running.
                api_octopart._set_from_env('enable', True, options, overwrite)
        if os.environ.get('KICOST_OCTOPART_EXTENDED'):
            api_octopart._set_from_env('extended', True, options, overwrite)


distributor_class.register(api_octopart, 60)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools
import os
import re
import kicost
import subprocess

try:
    from setuptools import setup
    from setuptools.command.develop import develop
    from setuptools.command.install import install
except ImportError:
    from distutils.core import setup
    from distutils.core.command.develop import develop
    from distutils.core.command.install import install

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def post_install_setup():
    # Run the KiCost integration script.
    try:
        import sys
        if sys.platform.startswith("win32"):
            # For Windows it is necessary one additional library (used to create the shortcut).
            print('Installing additional library need for Windows setup...')
            try:
                if sys.version_info < (3, 0):
                    from pip._internal import main as pipmain
                else:
                    from pip import main as pipmain
                pipmain(['install', 'pywin32'])
            except Exception:
                print('Error to install Windows additional Python library. KiCost configuration related to Windows registry may not work.')
        # Run setup: shortcut, BOM module to Eeschema and OS context menu.
        try:
            from .kicost.kicost_config import kicost_setup
            kicost_setup()
        except Exception:
            print('Running the external configuration command...')
            from subprocess import call
            call(['kicost', '--setup'])
    except Exception:
        print('Error to run KiCost integration script.')


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        post_install_setup
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        post_install_setup
        install.run(self)


# Update the information files that stay in the computer.
with open('README.rst') as readme_file:
    readme = readme_file.read()
readme = readme.replace('block_diag.png', 'https://raw.githubusercontent.com/hildogjr/KiCost/master/docs/block_diag.png')
readme = readme.replace('gui.png', 'https://raw.githubusercontent.com/hildogjr/KiCost/master/docs/gui.png')
readme = readme.replace('kitspace.png', 'https://raw.githubusercontent.com/hildogjr/KiCost/master/kicost/kitspace.png')
with open(os.path.join('kicost', 'HISTORY.rst')) as history_file:
    history_full = history_file.read()
    try:
        SHOW_LAST_HISTORY = 3
        RE_TITLE_SEPARATOR = r'(.|\n|\r|\_)'
        update_format = (r'(?P<last_history>'
                         r'History\s\-+\s' + RE_TITLE_SEPARATOR +
                         r'*?(' + RE_TITLE_SEPARATOR +
                         r'*?\s{2,}){'+str(SHOW_LAST_HISTORY)+'}'
                         r')')
        history_lastest = re.search(update_format, history_full).group('last_history').strip()
        if history_lastest:
            if SHOW_LAST_HISTORY == 1:
                history_lastest = history_lastest.replace('History\n-------', 'Latest update\n-------------')
            else:
                history_lastest = history_lastest.replace('History\n-------', 'Latest updates\n--------------')
            history = history_lastest + '\n\nAccess https://github.com/hildogjr/KiCost/blob/master/kicost/HISTORY.rst for full development history.'
        else:
            history = history_full
    except Exception:
        history = history_full
        pass

# KiCost Python packages requirements to run-time.
with open('requirements.txt') as f:
    requirements = f.read().splitlines()
# Remove the comments of the line.
for idx, r in enumerate(requirements):
    requirements[idx] = re.findall('^(.*)(!?#.*)*', r)[0][0].strip()
if '' in requirements:
    requirements.remove('')

# KiCost Python packages requirements to debug and tests.
test_requirements = [
    # Put package test requirements here.
]

# Extra files needed by KiCost.
data_files = [
    # ('kicost', ['kicost/kicost.ico']), # Icon to the user guide. Added via `MANIFEST.in`.
]

with open('kicost/version.py', 'rt') as f:
    version_py = f.read()
with open('kicost/version.py', 'wt') as f:
    if 'KICOST_RELEASE' in os.environ:
        res = 'release'
    else:
        try:
            res = subprocess.check_output(['git', 'log', '-1', '--pretty=format:%h-%as']).decode('ascii')
        except (subprocess.CalledProcessError, FileNotFoundError):
            res = 'unknown'
    f.write(re.sub("__build__ = '(.*)'", "__build__ = '{}'".format(res), version_py))
if 'KICOST_FAKE_VERSION' in os.environ:
    kicost.__version__ = os.environ['KICOST_FAKE_VERSION']

setup(
    name='kicost',
    version=kicost.__version__,
    description="Build cost spreadsheet for a KiCad project.",
    long_description=readme + '\n\n' + history,
    # long_description_content_type="text/reStructuredText",
    author=kicost.__author__,
    author_email=kicost.__email__,
    url='https://hildogjr.github.io/KiCost',
    project_urls={
        'Doc': 'https://hildogjr.github.io/KiCost',
        'Git': 'https://github.com/hildogjr/KiCost',
    },
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': ['kicost = kicost.__main__:main']},
    package_dir={'kicost': 'kicost'},
    include_package_data=True,
    package_data={'kicost': ['*.gif', '*.png', '*.txt']},
    # data_files=data_files,
    scripts=[],
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='KiCAD, BOM, electronics',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    }
)



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Libraries.
import argparse as ap  # Command argument parser.
import os
import sys
import platform
import time
import tqdm
import logging
import subprocess
# Debug, language and default configurations.
from .global_vars import wxPythonNotPresent, DEF_MAX_COLUMN_W, W_CMDLINE
# Sub systems can be imported before logger
from .edas import get_registered_eda_names, set_edas_logger  # noqa: E402
from .distributors import (get_distributors_list, set_distributors_logger, set_distributors_progress, is_valid_api,
                           init_distributor_dict, configure_from_environment, configure_apis)
# KiCost definitions and modules/packages functions.
from .kicost import output_filename, kicost_gui_notdependences, kicost  # kicost core functions.
try:
    from .kicost_gui import kicost_gui
    GUI_ENABLED = True
except wxPythonNotPresent:
    # If the wxPython dependences are not installed and the user just want the KiCost CLI.
    GUI_ENABLED = False
from .spreadsheet import Spreadsheet
from .config import load_config, config_force_ttl, config_force_path
from . import __version__, __build__, debug_obsessive, debug_detailed, error, warning, info, get_logger, set_logger, KiCostError
from . import log


def init_all_loggers(main_logger, dist_logger, edas_logger):
    set_logger(main_logger)
    set_distributors_logger(dist_logger)
    set_edas_logger(edas_logger)


# Python 2.7 compatibility
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

###############################################################################
# Additional functions
###############################################################################


def kicost_version_info():
    build = __build__
    # Are we running from a repo? (including "pip install -e")
    prev_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    if os.path.isdir(os.path.join(prev_dir, '.git')):
        cwd = os.getcwd()
        # Try to change to this place and find the git hash
        try:
            os.chdir(prev_dir)
            build = subprocess.check_output(['git', 'log', '-1', '--pretty=format:%h-%as']).decode('ascii')
        except (OSError, subprocess.CalledProcessError, FileNotFoundError):
            pass
        finally:
            os.chdir(cwd)
    version_info_str = r'KiCost v{} ({})'.format(__version__, build)
    version_info_str += r' at Python {}.{}.{}'.format(sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
    version_info_str += r' on {}({}). '.format(platform.platform(), platform.architecture()[0])
    try:
        import wx
        version_info_str += 'Graphical library: {}.'.format(wx.version())
    except ImportError:
        version_info_str += 'No graphical library installed for the GUI.'
    return version_info_str


class TqdmLoggingHandler(logging.Handler):
    '''Overload the class to write the logging through the `tqdm`.'''
    def __init__(self, stream, level=logging.NOTSET):
        super(self.__class__, self).__init__(level)
        self.stream = stream

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg, file=self.stream)  # , nolock=True)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


class ProgressConsole(object):
    def __init__(self, total, logger):
        # Create a handler that emits using TQDM
        self.logTqdmHandler = TqdmLoggingHandler(sys.stderr)
        # Apply our custom formatter
        self.logTqdmHandler.setFormatter(log.CustomFormatter(sys.stderr))
        # Add as a handler and avoid propagating to the base class
        logger.addHandler(self.logTqdmHandler)
        logger.propagate = False
        self.logger = logger
        # Create the progress object
        self.progress = tqdm.tqdm(desc='Progress', total=total, unit='part', miniters=1, file=sys.stderr)

    def update(self, val):
        self.progress.update(val)

    def close(self):
        self.progress.close()
        self.logger.removeHandler(self.logTqdmHandler)
        self.logger.propagate = True


def command_line_api_options(api_options, args):
    """ Applies the command line options that overrides the APIs config """
    # Force API options from the command line
    if args.cache_ttl is not None:
        config_force_ttl(args.cache_ttl)
    if args.cache_path is not None:
        config_force_path(args.cache_path)
    if args.octopart_key is not None:
        api_options['Octopart']['key'] = args.octopart_key
    if args.octopart_level is not None:
        level = args.octopart_level
        extended = False
        if level[-1] == 'p':
            extended = True
            level = level[:-1]
        api_options['Octopart']['level'] = level
        api_options['Octopart']['extended'] = extended
    for api in args.disable_api:
        debug_obsessive('Disabling API ' + api)
        if is_valid_api(api):
            api_options[api]['enable'] = False
        else:
            warning(W_CMDLINE, 'Unknown API `{}`'.format(api))
    for api in args.enable_api:
        debug_obsessive('Enabling API ' + api)
        if is_valid_api(api):
            api_options[api]['enable'] = True
        else:
            warning(W_CMDLINE, 'Unknown API `{}`'.format(api))
    debug_detailed('Final API options {}'.format(api_options))


def configure_kicost_apis(config_file, overwrite, apply_user_opts, data):
    # Configuration file
    api_options = load_config(config_file)
    # Environment, overwrite only if the config file wasn't specified in the command line
    configure_from_environment(api_options, overwrite)
    # Apply command line options
    apply_user_opts(api_options, data)
    # Configure the APIs
    configure_apis(api_options)
    # Start with a clean list of available distributors
    init_distributor_dict()


###############################################################################
# Command-line interface.
###############################################################################


def main_real():

    parser = ap.ArgumentParser(
        description='Build cost spreadsheet for a KiCAD project.')
    parser.add_argument('-v', '--version',
                        action='version',
                        version='KiCost v{}'.format(__version__))
    parser.add_argument('--info',
                        action='version',
                        version=kicost_version_info(),
                        help='Show program\' and library information and version.')
    parser.add_argument('-i', '--input',
                        nargs='+',
                        type=str,
                        metavar='FILE.XML',
                        help='One or more schematic BOM XML files.')
    parser.add_argument('-o', '--output',
                        nargs='?',
                        type=str,
                        metavar='FILE.XLSX',
                        help='Generated cost spreadsheet.')
    parser.add_argument('-f', '--fields',
                        nargs='+',
                        type=str,
                        default=[],
                        metavar='NAME',
                        help='''Specify the names of additional part fields to
                            extract and insert in the global data section of
                            the spreadsheet.''')
    parser.add_argument('--translate_fields',
                        nargs='+',
                        type=str,
                        metavar='NAME',
                        help='''Specify or remove field translation
                            (--translate X1 Y1 X2 Y2 X3 ~,
                            translates X1 to Y1 and X2 to Y2 and remove
                            X3 for the internal dictionary).''')
    parser.add_argument('--variant',
                        nargs='+',
                        type=str,
                        default=[' '],  # Default variant is a space.
                        help='schematic variant name filter using regular expression.')
    parser.add_argument('-w', '--overwrite',
                        action='store_true',
                        help='Allow overwriting of an existing spreadsheet.')
    parser.add_argument('-q', '--quiet',
                        action='store_true',
                        help='Enable quiet mode with no warnings.')
    parser.add_argument('--ignore_fields',
                        nargs='+',
                        default=[],
                        help='Declare part fields to ignore when reading the BoM file.',
                        metavar='NAME',
                        type=str)
    parser.add_argument('--group_fields',
                        nargs='+',
                        default=[],
                        help='Declare part fields to merge when grouping parts.',
                        metavar='NAME',
                        type=str)
    parser.add_argument('--split_extra_fields',
                        nargs='+',
                        default=[],
                        help='Declare part fields to include in multipart split process.',
                        metavar='NAME',
                        type=str)
    parser.add_argument('--debug',
                        nargs='?',
                        type=int,
                        default=None,
                        metavar='LEVEL',
                        help='Print debugging info. (Larger LEVEL means more info.)')
    parser.add_argument('--eda', choices=get_registered_eda_names(),
                        nargs='+',
                        default=['kicad'],
                        help='Choose EDA tool from which the XML BOM file originated, or use csv for .CSV files.')
    parser.add_argument('--show_dist_list',
                        action='store_true',
                        help='Show list of distributors that can be scraped for cost data, then exit.')
    parser.add_argument('--show_eda_list',
                        action='store_true',
                        help='Show list of EDA tools whose files KiCost can read, then exit.')
    parser.add_argument('--no_collapse',
                        action='store_true',
                        help='Do not collapse the part references in the spreadsheet.')
    parser.add_argument('--show_cat_url',
                        action='store_true',
                        help='Do not suppress the catalogue links into the catalogue code in the spreadsheet.')
    parser.add_argument('-e', '--exclude',
                        nargs='+', type=str, default=[],
                        metavar='DIST',
                        help='Excludes the given distributor(s) from the scraping process.')
    parser.add_argument('--include',
                        nargs='+', type=str, default=[],
                        metavar='DIST',
                        help='Includes only the given distributor(s) in the scraping process.')
    parser.add_argument('--no_price',
                        action='store_true',
                        help='Create a spreadsheet without scraping part data from distributor websites.')
    parser.add_argument('--currency',
                        nargs='?',
                        type=str,
                        default='USD',
                        help='Define the priority currency. Use the ISO4217 for currency (`USD`, `EUR`). Default: `USD`.')
    parser.add_argument('-n', '--board_qty',
                        nargs='+', type=int, default=[Spreadsheet.DEFAULT_BUILD_QTY], metavar='QTY',
                        help='Number of boards to fabricate. Default: ' + str(Spreadsheet.DEFAULT_BUILD_QTY) + '.')
    parser.add_argument('--max_column_width',
                        nargs='?',
                        type=int,
                        default=DEF_MAX_COLUMN_W,
                        metavar='WIDTH',
                        help='Maximum column width. Using 0 disables the cell size adjust. Default: ' + str(DEF_MAX_COLUMN_W) + '.')
    parser.add_argument('--gui',
                        nargs='+',
                        type=str,
                        metavar='FILE.XML',
                        help='Start the GUI to run KiCost passing the file parameter give by "--input",'
                             ' all others parameters are ignored.')
    # SET: This is half imlemented. Not currently working.
    #     parser.add_argument('--user',
    #                         action='store_true',
    #                         help='Run KiCost on terminal using the parameters in the GUI memory, all passed parameters from'
    #                              ' terminal take priority.')
    parser.add_argument('--force_en_us',
                        action='store_true',
                        help='Workaround for broken wxWidgets locale on some Windows systems.')
    parser.add_argument('--setup',
                        action='store_true',
                        help='Run KiCost integration (with KiCad and OS) configuration script.')
    parser.add_argument('--unsetup',
                        action='store_true',
                        help='Undo the KiCost integration.')
    parser.add_argument('--octopart_key',
                        nargs='?', type=str, metavar='APIKEY',
                        help='Enable Octopart using the provided key. Use None to disable it.')
    parser.add_argument('--octopart_level',
                        nargs='?', type=str, metavar='APILEVEL', choices=['3', '3p', '4', '4p'],
                        help='Use Octopart API level. Can be 3 or 4 for basic API and 3p or 4p for PRO plans. Default: 4')
    parser.add_argument('--disable_api',
                        nargs='+', type=str, default=[],
                        metavar='API',
                        help='Disable the listed APIs.')
    parser.add_argument('--enable_api',
                        nargs='+', type=str, default=[],
                        metavar='API',
                        help='Enable the listed APIs.')
    parser.add_argument('-c', '--config',
                        nargs='?', type=str,
                        metavar='CONFIG.YAML',
                        help='Configuration file.')
    parser.add_argument('--cache_ttl',
                        nargs='?', type=float,
                        metavar='DAYS',
                        help='Force the Time To Live for the APIs caches.')
    parser.add_argument('--cache_path',
                        nargs='?', type=str,
                        metavar='BASE_PATH',
                        help='Force the base path for the APIs caches.')

    args = parser.parse_args()

    # Set up logging verbosity
    log.set_domain('kicost')
    init_all_loggers(log.init(), log.get_logger('kicost.distributors'), log.get_logger('kicost.edas'))
    log.set_verbosity(get_logger(), args.debug, args.quiet)
    set_distributors_progress(ProgressConsole)

    # Now we can print errors, inform if YAML is missing
    try:
        import yaml  # noqa F401
    except ImportError:
        error('No yaml module for Python, install it (i.e. python3-yaml)')

    # APIs Configuration
    configure_kicost_apis(args.config, args.config is None, command_line_api_options, args)

    # Setup and unsetup KiCost integration.
    if args.setup:
        from .kicost_config import kicost_setup
        kicost_setup()
        return
    if args.unsetup:
        from .kicost_config import kicost_unsetup
        kicost_unsetup()
        return

    # Simple commands
    if args.show_dist_list:
        info('Distributor list: ' + ' '.join(sorted(get_distributors_list())))
        sys.exit(0)
    if args.show_eda_list:
        info('EDA supported list: ' + ' '.join(sorted(get_registered_eda_names())))
        sys.exit(0)

    # Set up spreadsheet output file.
    if args.output is None:
        # If no output file is given...
        if args.input is not None:
            # Send output to spreadsheet with name of input file.
            # Compose a name with the multiple BOM input file names.
            args.output = output_filename(args.input)
        else:
            # Send output to spreadsheet with name of this application.
            args.output = os.path.splitext(sys.argv[0])[0] + '.xlsx'
    else:
        # Output file was given. Make sure it has spreadsheet extension.
        args.output = os.path.splitext(args.output)[0] + '.xlsx'

    # SET: This is half implemented. Not currently working.
    #     # Call the KiCost interface to alredy run KiCost, this is just to use the
    #     # saved user configurations of the graphical interface.
    #     if args.user:
    #         if not GUI_ENABLED:
    #             kicost_gui_notdependences()
    #         kicost_gui_runterminal(args)
    #         sys.exit(0)

    # Handle case where output is going into an existing spreadsheet file.
    if os.path.isfile(args.output):
        if not args.overwrite:
            error('Output file {} already exists!\nUse the --overwrite option to replace it.'.format(args.output))
            sys.exit(2)

    if args.gui:
        if not GUI_ENABLED:
            kicost_gui_notdependences()
        kicost_gui(args.force_en_us, [os.path.abspath(fileName) for fileName in args.gui])
        sys.exit(0)

    if args.input is None:
        if not GUI_ENABLED:
            kicost_gui_notdependences()
        kicost_gui(args.force_en_us)  # Use the user gui if no input is given.
        sys.exit(0)
    else:
        # Match the EDA tool formats with the input files.
        if len(args.eda) == 1:
            # Expand a single EDA format to cover all input files.
            args.eda = args.eda[0:1] * len(args.input)
        if len(args.input) != len(args.eda):
            error('The number of input files must match the number of EDA tool formats.')
            sys.exit(2)

        # Match the variants with the input files.
        if len(args.variant) == 1:
            args.variant = args.variant[0:1] * len(args.input)
        if len(args.input) != len(args.variant):
            error('The number of input files must match the number of variants.')
            sys.exit(2)

        # Match the board quantities with the input files.
        if len(args.board_qty) == 1:
            args.board_qty = args.board_qty[0:1] * len(args.input)
        if len(args.input) != len(args.board_qty):
            error('The number of input files must match the number of board quantities.')
            sys.exit(2)

        # Otherwise get XML from the given file.
        for i in range(len(args.input)):
            # Set '.xml' as the default file extension, treating this exception
            # allow other files extension and formats.
            try:
                if os.path.splitext(args.input[i])[1] == '':
                    args.input[i] += '.xml'
                elif os.path.splitext(args.input[i])[1] == '.csv':
                    args.eda[i] = 'csv'
            except IndexError:
                pass

    # List of distributors to scrape
    available = get_distributors_list()
    for d in args.include + args.exclude:
        if d not in available:
            error('Unknown distributor requested: `{}`'.format(d))
            sys.exit(2)
    if args.no_price:
        # None
        dist_list = []
    else:
        if not args.include:
            # All by default
            dist_list = available
        else:
            # Requested to be included
            dist_list = args.include
        # Requested to be excluded
        for d in args.exclude:
            dist_list.remove(d)

    debug_obsessive('Started ' + kicost_version_info())

    kicost(in_file=args.input, eda_name=args.eda,
           out_filename=args.output, collapse_refs=not args.no_collapse, suppress_cat_url=not args.show_cat_url,
           user_fields=args.fields, ignore_fields=args.ignore_fields,
           group_fields=args.group_fields, translate_fields=args.translate_fields,
           variant=args.variant, dist_list=dist_list, currency=args.currency, max_column_width=args.max_column_width,
           split_extra_fields=args.split_extra_fields, board_qty=args.board_qty)


def main():
    try:
        main_real()
    except KiCostError as e:
        error(e.msg)
        sys.exit(e.id)


###############################################################################
# Main entrypoint.
###############################################################################
if __name__ == '__main__':
    start_time = time.time()
    main()
    debug_obsessive('Elapsed time: %f seconds', time.time() - start_time)
